""" Defines common flow processing functionality
"""
import sys
import json

from ovs_dbg.ofp import OFPFlow
from ovs_dbg.decoders import FlowEncoder
from ovs_dbg.ofparse.console import ConsoleFormatter, print_context


def process_flows(flow_factory, callback, filename="", filter=None):
    """Process flows from file or stdin

    Args:
        flow_factory(Callable): function to call to create a flow
        callback (Callable): function to call with each processed flow
        filename (str): Optional; filename to read frows from
        filter (OFFilter): Optional; filter to use to filter flows
    """
    idx = 0
    if filename:
        with open(filename) as f:
            for line in f:
                flow = flow_factory(line, idx)
                idx += 1
                if not flow or (filter and not filter.evaluate(flow)):
                    continue
                callback(flow)
    else:
        data = sys.stdin.read()
        for line in data.split("\n"):
            line = line.strip()
            if line:
                flow = flow_factory(line, idx)
                idx += 1
                if not flow or (filter and not filter.evaluate(flow)):
                    continue
                callback(flow)


def tojson(flow_factory, opts):
    """
    Print the json representation of the flow list

    Args:
        flow_factory (Callable): Function to call to create the flows
        opts (dict): Options
    """
    flows = []

    def callback(flow):
        flows.append(flow)

    process_flows(flow_factory, callback, opts.get("filename"), opts.get("filter"))

    flow_json = json.dumps(
        [flow.dict() for flow in flows],
        indent=4,
        cls=FlowEncoder,
    )

    print(flow_json)


def pprint(flow_factory, opts):
    """
    Pretty print the flows

    Args:
        flow_factory (Callable): Function to call to create the flows
        opts (dict): Options
    """
    console = ConsoleFormatter(opts)

    def callback(flow):
        high = None
        if opts.get("highlight"):
            result = opts.get("highlight").evaluate(flow)
            if result:
                high = result.kv
        console.print_flow(flow, high)

    with print_context(console.console, opts):
        process_flows(flow_factory, callback, opts.get("filename"), opts.get("filter"))



import click
import sys
import configparser

from ovs_dbg.filter import OFFilter
from pkg_resources import resource_filename

_default_config_file = "ofparse.conf"
_default_config_path = resource_filename(__name__, _default_config_file)


class Options(dict):
    """Options dictionary"""

    pass


@click.group(
    subcommand_metavar="TYPE", context_settings=dict(help_option_names=["-h", "--help"])
)
@click.option(
    "-c",
    "--config",
    help="Use non-default config file",
    type=click.Path(),
)
@click.option(
    "--style",
    help="Select style (defined in config file)",
    default=None,
    show_default=True,
)
@click.option(
    "-i",
    "-input",
    "filename",
    help="Read flows from specified filepath. If not provided, flows will be"
    " read from stdin",
    type=click.Path(),
)
@click.option(
    "-p",
    "--paged",
    help="Page the result (uses $PAGER). If colors are not disabled you might "
    'need to enable colors on your PAGER, eg: export PAGER="less -r".',
    is_flag=True,
    default=False,
    show_default=True,
)
@click.option(
    "-f",
    "--filter",
    help="Filter flows that match the filter expression. Run 'ofparse filter'"
    "for a detailed description of the filtering syntax",
    type=str,
    show_default=False,
)
@click.option(
    "-l",
    "--highlight",
    help="Highlight flows that match the filter expression. Run 'ofparse filter'"
    "for a detailed description of the filtering syntax",
    type=str,
    show_default=False,
)
@click.pass_context
def maincli(ctx, config, style, filename, paged, filter, highlight):
    """
    OpenFlow Parse utility.

    It parses openflow flows (such as the output of ovs-ofctl 'dump-flows') and
    prints them in different formats.

    """
    ctx.obj = Options()
    ctx.obj["filename"] = filename or ""
    ctx.obj["paged"] = paged
    if filter:
        try:
            ctx.obj["filter"] = OFFilter(filter)
        except Exception as e:
            raise click.BadParameter("Wrong filter syntax: {}".format(e))

    if highlight:
        try:
            ctx.obj["highlight"] = OFFilter(highlight)
        except Exception as e:
            raise click.BadParameter("Wrong filter syntax: {}".format(e))

    config_file = config or _default_config_path
    parser = configparser.ConfigParser()
    parser.read(config_file)

    ctx.obj["config"] = parser
    ctx.obj["style"] = style


@maincli.command(hidden=True)
@click.pass_context
def filter(ctx):
    """
    \b
    Filter Syntax
    *************

        [! | not ] {key}[[.subkey[.subkey]..] [= | > | < | ~=] {value})] [&& | || | or | and | not ] ...

    \b
    Comparison operators are:
        =   equality
        <   less than
        >   more than
        ~=  masking (valid for IP and Ethernet fields)

    \b
    Logical operators are:
        !{expr}:  NOT
        {expr} && {expr}: AND
        {expr} || {expr}: OR

    \b
    Matches and flow metadata:
        To compare against a match or info field, use the field directly, e.g:
            priority=100
            n_bytes>10
        Use simple keywords for flags:
            tcp and ip_src=192.168.1.1
    \b
    Actions:
        Actions values might be dictionaries, use subkeys to access individual
        values, e.g:
            output.port=3
        Use simple keywords for flags
            drop

    \b
    Examples of valid filters.
        nw_addr~=192.168.1.1 && (tcp_dst=80 || tcp_dst=443)
        arp=true && !arp_tsa=192.168.1.1
        n_bytes>0 && drop=true"""
    click.echo(ctx.command.get_help(ctx))


def main():
    """
    Main Function
    """
    maincli()



import re
import functools

from ovs_dbg.kv import KeyValue, KeyMetadata, ParseError
from ovs_dbg.decoders import decode_default


class ListDecoders:
    """ListDecoders is used by ListParser to decode the elements in the list

    A decoder is a function that accepts a value and returns its decoded
    object
    The list_decoder to be used is determined by index in the list provided to
    ListDecoders is important.

    Args:
        decoders (list of tuples): Optional,  A list of tuples.
            The first element in the tuple is the keyword associated with the
            value. The second element in the tuple is the decoder function.
    """

    def __init__(self, decoders=None):
        self._decoders = decoders or list()

    def decode(self, index, value_str):
        """Decode the index'th element of the list

        Args:
            index (int): the position in the list of the element ot decode
            value_str (str): the value string to decode
        """
        if index < 0 or index >= len(self._decoders):
            return self._default_decoder(index, value_str)

        try:
            key = self._decoders[index][0]
            value = self._decoders[index][1](value_str)
            return key, value
        except Exception as e:
            raise ParseError("Failed to decode value_str %s: %s" % (value_str, str(e)))

    @staticmethod
    def _default_decoder(index, value):
        key = "elem_{}".format(index)
        return key, decode_default(value)


class ListParser:
    """ListParser parses a list of values and stores them as key-value pairs

    It uses a ListDecoders instance to decode each element in the list.

    Args:
        decoders (ListDecoders): Optional, the decoders to use
        delims (list): Optional, list of delimiters of the list. Defaults to
            [',']
    """

    def __init__(self, decoders=None, delims=None):
        self._decoders = decoders or ListDecoders()
        self._keyval = list()
        delims = delims or [","]
        delims.append("$")
        self._regexp = r"({})".format("|".join(delims))

    def kv(self):
        return self._keyval

    def __iter__(self):
        return iter(self._keyval)

    def parse(self, string):
        """Parse the list in string

        Args:
            string (str): the string to parse

        Raises:
            ParseError if any parsing error occurs.
        """
        kpos = 0
        index = 0
        while kpos < len(string) and string[kpos] != "\n":
            split_parts = re.split(self._regexp, string[kpos:], 1)
            if len(split_parts) < 3:
                break

            value_str = split_parts[0]

            key, value = self._decoders.decode(index, value_str)

            meta = KeyMetadata(
                kpos=kpos,
                vpos=kpos,
                kstring=value_str,
                vstring=value_str,
            )
            self._keyval.append(KeyValue(key, value, meta))

            kpos += len(value_str) + 1
            index += 1


def decode_nested_list(decoders, delims, value):
    """Extracts nested list from te string and returns it in a dictionary
    them in a dictionary

    Args:
        decoders (ListDecoders): the ListDecoders to use.
        value (str): the value string to decode.
    """
    parser = ListParser(decoders, delims)
    parser.parse(value)
    return {kv.key: kv.value for kv in parser.kv()}


def nested_list_decoder(decoders=None, delims=None):
    """Helper function that creates a nested list decoder with given
    ListDecoders
    """
    return functools.partial(decode_nested_list, decoders, delims)



# This file is auto-generated. Do not edit

import functools
from ovs_dbg import decoders

field_decoders = {
    "dp_hash": decoders.Mask32,
    "recirc_id": decoders.decode_int,
    "packet_type": decoders.decode_default,
    "conj_id": decoders.decode_int,
    "tun_id": decoders.Mask64,
    "tunnel_id": decoders.Mask64,
    "tun_src": decoders.IPMask,
    "tun_dst": decoders.IPMask,
    "tun_ipv6_src": decoders.IPMask,
    "tun_ipv6_dst": decoders.IPMask,
    "tun_flags": decoders.decode_default,
    "tun_ttl": decoders.decode_int,
    "tun_tos": decoders.decode_int,
    "tun_gbp_id": decoders.Mask16,
    "tun_gbp_flags": decoders.Mask8,
    "tun_erspan_idx": decoders.decode_mask(20),
    "tun_erspan_ver": decoders.decode_mask(4),
    "tun_erspan_dir": decoders.decode_mask(1),
    "tun_erspan_hwid": decoders.decode_mask(6),
    "tun_gtpu_flags": decoders.Mask8,
    "tun_gtpu_msgtype": decoders.Mask8,
    "tun_metadata0": decoders.Mask992,
    "tun_metadata1": decoders.Mask992,
    "tun_metadata2": decoders.Mask992,
    "tun_metadata3": decoders.Mask992,
    "tun_metadata4": decoders.Mask992,
    "tun_metadata5": decoders.Mask992,
    "tun_metadata6": decoders.Mask992,
    "tun_metadata7": decoders.Mask992,
    "tun_metadata8": decoders.Mask992,
    "tun_metadata9": decoders.Mask992,
    "tun_metadata10": decoders.Mask992,
    "tun_metadata11": decoders.Mask992,
    "tun_metadata12": decoders.Mask992,
    "tun_metadata13": decoders.Mask992,
    "tun_metadata14": decoders.Mask992,
    "tun_metadata15": decoders.Mask992,
    "tun_metadata16": decoders.Mask992,
    "tun_metadata17": decoders.Mask992,
    "tun_metadata18": decoders.Mask992,
    "tun_metadata19": decoders.Mask992,
    "tun_metadata20": decoders.Mask992,
    "tun_metadata21": decoders.Mask992,
    "tun_metadata22": decoders.Mask992,
    "tun_metadata23": decoders.Mask992,
    "tun_metadata24": decoders.Mask992,
    "tun_metadata25": decoders.Mask992,
    "tun_metadata26": decoders.Mask992,
    "tun_metadata27": decoders.Mask992,
    "tun_metadata28": decoders.Mask992,
    "tun_metadata29": decoders.Mask992,
    "tun_metadata30": decoders.Mask992,
    "tun_metadata31": decoders.Mask992,
    "tun_metadata32": decoders.Mask992,
    "tun_metadata33": decoders.Mask992,
    "tun_metadata34": decoders.Mask992,
    "tun_metadata35": decoders.Mask992,
    "tun_metadata36": decoders.Mask992,
    "tun_metadata37": decoders.Mask992,
    "tun_metadata38": decoders.Mask992,
    "tun_metadata39": decoders.Mask992,
    "tun_metadata40": decoders.Mask992,
    "tun_metadata41": decoders.Mask992,
    "tun_metadata42": decoders.Mask992,
    "tun_metadata43": decoders.Mask992,
    "tun_metadata44": decoders.Mask992,
    "tun_metadata45": decoders.Mask992,
    "tun_metadata46": decoders.Mask992,
    "tun_metadata47": decoders.Mask992,
    "tun_metadata48": decoders.Mask992,
    "tun_metadata49": decoders.Mask992,
    "tun_metadata50": decoders.Mask992,
    "tun_metadata51": decoders.Mask992,
    "tun_metadata52": decoders.Mask992,
    "tun_metadata53": decoders.Mask992,
    "tun_metadata54": decoders.Mask992,
    "tun_metadata55": decoders.Mask992,
    "tun_metadata56": decoders.Mask992,
    "tun_metadata57": decoders.Mask992,
    "tun_metadata58": decoders.Mask992,
    "tun_metadata59": decoders.Mask992,
    "tun_metadata60": decoders.Mask992,
    "tun_metadata61": decoders.Mask992,
    "tun_metadata62": decoders.Mask992,
    "tun_metadata63": decoders.Mask992,
    "metadata": decoders.Mask64,
    "in_port": decoders.decode_default,
    "in_port_oxm": decoders.decode_default,
    "actset_output": decoders.decode_default,
    "skb_priority": decoders.decode_int,
    "pkt_mark": decoders.Mask32,
    "ct_state": decoders.decode_default,
    "ct_zone": decoders.decode_int,
    "ct_mark": decoders.Mask32,
    "ct_label": decoders.Mask128,
    "ct_nw_proto": decoders.decode_int,
    "ct_nw_src": decoders.IPMask,
    "ct_nw_dst": decoders.IPMask,
    "ct_ipv6_src": decoders.IPMask,
    "ct_ipv6_dst": decoders.IPMask,
    "ct_tp_src": decoders.Mask16,
    "ct_tp_dst": decoders.Mask16,
    "reg0": decoders.Mask32,
    "reg1": decoders.Mask32,
    "reg2": decoders.Mask32,
    "reg3": decoders.Mask32,
    "reg4": decoders.Mask32,
    "reg5": decoders.Mask32,
    "reg6": decoders.Mask32,
    "reg7": decoders.Mask32,
    "reg8": decoders.Mask32,
    "reg9": decoders.Mask32,
    "reg10": decoders.Mask32,
    "reg11": decoders.Mask32,
    "reg12": decoders.Mask32,
    "reg13": decoders.Mask32,
    "reg14": decoders.Mask32,
    "reg15": decoders.Mask32,
    "xreg0": decoders.Mask64,
    "xreg1": decoders.Mask64,
    "xreg2": decoders.Mask64,
    "xreg3": decoders.Mask64,
    "xreg4": decoders.Mask64,
    "xreg5": decoders.Mask64,
    "xreg6": decoders.Mask64,
    "xreg7": decoders.Mask64,
    "xxreg0": decoders.Mask128,
    "xxreg1": decoders.Mask128,
    "xxreg2": decoders.Mask128,
    "xxreg3": decoders.Mask128,
    "eth_src": decoders.EthMask,
    "dl_src": decoders.EthMask,
    "eth_dst": decoders.EthMask,
    "dl_dst": decoders.EthMask,
    "eth_type": decoders.decode_int,
    "dl_type": decoders.decode_int,
    "vlan_tci": decoders.Mask16,
    "dl_vlan": decoders.decode_int,
    "vlan_vid": decoders.decode_mask(12),
    "dl_vlan_pcp": decoders.decode_int,
    "vlan_pcp": decoders.decode_int,
    "mpls_label": decoders.decode_int,
    "mpls_tc": decoders.decode_int,
    "mpls_bos": decoders.decode_int,
    "mpls_ttl": decoders.decode_int,
    "ip_src": decoders.IPMask,
    "nw_src": decoders.IPMask,
    "ip_dst": decoders.IPMask,
    "nw_dst": decoders.IPMask,
    "ipv6_src": decoders.IPMask,
    "ipv6_dst": decoders.IPMask,
    "ipv6_label": decoders.decode_mask(20),
    "nw_proto": decoders.decode_int,
    "ip_proto": decoders.decode_int,
    "nw_tos": decoders.decode_int,
    "ip_dscp": decoders.decode_int,
    "nw_ecn": decoders.decode_int,
    "ip_ecn": decoders.decode_int,
    "nw_ttl": decoders.decode_int,
    "ip_frag": decoders.decode_default,
    "nw_frag": decoders.decode_default,
    "arp_op": decoders.decode_int,
    "arp_spa": decoders.IPMask,
    "arp_tpa": decoders.IPMask,
    "arp_sha": decoders.EthMask,
    "arp_tha": decoders.EthMask,
    "tcp_src": decoders.Mask16,
    "tp_src": decoders.Mask16,
    "tcp_dst": decoders.Mask16,
    "tp_dst": decoders.Mask16,
    "tcp_flags": decoders.decode_default,
    "udp_src": decoders.Mask16,
    "udp_dst": decoders.Mask16,
    "sctp_src": decoders.Mask16,
    "sctp_dst": decoders.Mask16,
    "icmp_type": decoders.decode_int,
    "icmp_code": decoders.decode_int,
    "icmpv6_type": decoders.decode_int,
    "icmpv6_code": decoders.decode_int,
    "nd_target": decoders.IPMask,
    "nd_sll": decoders.EthMask,
    "nd_tll": decoders.EthMask,
    "nd_reserved": decoders.decode_int,
    "nd_options_type": decoders.decode_int,
    "nsh_flags": decoders.Mask8,
    "nsh_mdtype": decoders.decode_int,
    "nsh_np": decoders.decode_int,
    "nsh_spi": decoders.decode_int,
    "nsp": decoders.decode_int,
    "nsh_si": decoders.decode_int,
    "nsi": decoders.decode_int,
    "nsh_c1": decoders.Mask32,
    "nshc1": decoders.Mask32,
    "nsh_c2": decoders.Mask32,
    "nshc2": decoders.Mask32,
    "nsh_c3": decoders.Mask32,
    "nshc3": decoders.Mask32,
    "nsh_c4": decoders.Mask32,
    "nshc4": decoders.Mask32,
    "nsh_ttl": decoders.decode_int,
}



""" Defines the parsers needed to parse ofproto flows
"""

from dataclasses import dataclass
import functools
import re

from ovs_dbg.kv import KVParser, KVDecoders, ParseError, nested_kv_decoder
from ovs_dbg.fields import field_decoders
from ovs_dbg.flow import Flow, Section
from ovs_dbg.list import ListDecoders, nested_list_decoder
from ovs_dbg.decoders import (
    decode_default,
    decode_flag,
    decode_int,
    decode_time,
    decode_mask,
    IPMask,
    EthMask,
    decode_free_output,
    decode_nat,
)
from ovs_dbg.ofp_act import (
    decode_output,
    decode_field,
    decode_controller,
    decode_bundle,
    decode_bundle_load,
    decode_encap_ethernet,
    decode_load_field,
    decode_set_field,
    decode_move_field,
    decode_dec_ttl,
    decode_chk_pkt_larger,
    decode_zone,
    decode_exec,
    decode_learn,
)


class OFPFlow(Flow):
    """OpenFlow Flow"""

    def __init__(self, sections, orig="", id=None):
        """Constructor"""
        super(OFPFlow, self).__init__(sections, orig, id)

    @classmethod
    def from_string(cls, ofp_string, id=None):
        """Parse a ofproto flow string

        The string is expected to have the follwoing format:
            [flow data] [match] actions=[actions]

        :param ofp_string: a ofproto string as dumped by ovs-ofctl tool
        :type ofp_string: str

        :return: an OFPFlow with the content of the flow string
        :rtype: OFPFlow
        """
        sections = list()
        parts = ofp_string.split("actions=")
        if len(parts) != 2:
            raise ValueError("malformed ofproto flow: {}", ofp_string)

        actions = parts[1]

        field_parts = parts[0].rstrip(" ").rpartition(" ")
        if len(field_parts) != 3:
            raise ValueError("malformed ofproto flow: {}", ofp_string)

        info = field_parts[0]
        match = field_parts[2]

        info_decoders = cls._info_decoders()
        iparser = KVParser(info_decoders)
        iparser.parse(info)
        isection = Section(
            name="info", pos=ofp_string.find(info), string=info, data=iparser.kv()
        )
        sections.append(isection)

        match_decoders = KVDecoders(
            {**cls._field_decoders(), **cls._flow_match_decoders()}
        )
        mparser = KVParser(match_decoders)
        mparser.parse(match)
        msection = Section(
            name="match", pos=ofp_string.find(match), string=match, data=mparser.kv()
        )
        sections.append(msection)

        act_decoders = cls._act_decoders()
        aparser = KVParser(act_decoders)
        aparser.parse(actions)
        asection = Section(
            name="actions",
            pos=ofp_string.find(actions),
            string=actions,
            data=aparser.kv(),
            is_list=True,
        )
        sections.append(asection)

        return cls(sections, ofp_string, id)

    @classmethod
    def _info_decoders(cls):
        """Generate the match decoders"""
        info = {
            "table": decode_int,
            "duration": decode_time,
            "n_packet": decode_int,
            "n_bytes": decode_int,
            "cookie": decode_int,
            "idle_timeout": decode_time,
            "hard_timeout": decode_time,
            "hard_age": decode_time,
        }
        return KVDecoders(info)

    @classmethod
    def _flow_match_decoders(cls):
        """Returns the decoders for key-values that are part of the flow match
        but not a flow field"""
        return {
            "priority": decode_int,
        }

    @classmethod
    def _field_decoders(cls):
        shorthands = [
            "eth",
            "ip",
            "ipv6",
            "icmp",
            "icmp6",
            "tcp",
            "tcp6",
            "udp",
            "udp6",
            "sctp",
            "arp",
            "rarp",
            "mpls",
            "mplsm",
        ]

        return {**field_decoders, **{key: decode_flag for key in shorthands}}

    @classmethod
    def _output_actions_decoders(cls):
        """Returns the decoders for the output actions"""
        return {
            "output": decode_output,
            "drop": decode_flag,
            "controller": decode_controller,
            "enqueue": nested_list_decoder(
                ListDecoders([("port", decode_default), ("queue", int)]),
                delims=[",", ":"],
            ),
            "bundle": decode_bundle,
            "bundle_load": decode_bundle_load,
            "group": decode_default,
        }

    @classmethod
    def _encap_actions_decoders(cls):
        """Returns the decoders for the encap actions"""

        return {
            "pop_vlan": decode_flag,
            "strip_vlan": decode_flag,
            "push_vlan": decode_default,
            "decap": decode_flag,
            "encap": nested_kv_decoder(
                KVDecoders(
                    {
                        "nsh": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "md_type": decode_default,
                                    "tlv": nested_list_decoder(
                                        ListDecoders(
                                            [
                                                ("class", decode_int),
                                                ("type", decode_int),
                                                ("value", decode_int),
                                            ]
                                        )
                                    ),
                                }
                            )
                        ),
                    },
                    default=None,
                    default_free=decode_encap_ethernet,
                )
            ),
        }

    @classmethod
    def _field_action_decoders(cls):
        """Returns the decoders for the field modification actions"""
        # Field modification actions
        field_default_decoders = [
            "set_mpls_label",
            "set_mpls_tc",
            "set_mpls_ttl",
            "mod_nw_tos",
            "mod_nw_ecn",
            "mod_tcp_src",
            "mod_tcp_dst",
        ]
        return {
            "load": decode_load_field,
            "set_field": functools.partial(
                decode_set_field, KVDecoders(cls._field_decoders())
            ),
            "move": decode_move_field,
            "mod_dl_dst": EthMask,
            "mod_dl_src": EthMask,
            "mod_nw_dst": IPMask,
            "mod_nw_src": IPMask,
            "dec_ttl": decode_dec_ttl,
            "dec_mpls_ttl": decode_flag,
            "dec_nsh_ttl": decode_flag,
            "check_pkt_larger": decode_chk_pkt_larger,
            **{field: decode_default for field in field_default_decoders},
        }

    @classmethod
    def _meta_action_decoders(cls):
        """Returns the decoders for the metadata actions"""
        meta_default_decoders = ["set_tunnel", "set_tunnel64", "set_queue"]
        return {
            "pop_queue": decode_flag,
            **{field: decode_default for field in meta_default_decoders},
        }

    @classmethod
    def _fw_action_decoders(cls):
        """Returns the decoders for the Firewalling actions"""
        return {
            "ct": nested_kv_decoder(
                KVDecoders(
                    {
                        "commit": decode_flag,
                        "zone": decode_zone,
                        "table": decode_int,
                        "nat": decode_nat,
                        "force": decode_flag,
                        "exec": functools.partial(
                            decode_exec,
                            KVDecoders(
                                {
                                    **cls._encap_actions_decoders(),
                                    **cls._field_action_decoders(),
                                    **cls._meta_action_decoders(),
                                }
                            ),
                        ),
                        "alg": decode_default,
                    }
                )
            ),
            "ct_clear": decode_flag,
        }

    @classmethod
    def _control_action_decoders(cls):
        return {
            "resubmit": nested_list_decoder(
                ListDecoders(
                    [
                        ("port", decode_default),
                        ("table", decode_int),
                        ("ct", decode_flag),
                    ]
                )
            ),
            "push": decode_field,
            "pop": decode_field,
            "exit": decode_flag,
            "multipath": nested_list_decoder(
                ListDecoders(
                    [
                        ("fields", decode_default),
                        ("basis", decode_int),
                        ("algorithm", decode_default),
                        ("n_links", decode_int),
                        ("arg", decode_int),
                        ("dst", decode_field),
                    ]
                )
            ),
        }

    @classmethod
    def _clone_actions_decoders(cls, action_decoders):
        """Generate the decoders for clone actions

        Args:
            action_decoders (dict): The decoders of the supported nested actions
        """
        return {
            "learn": decode_learn(
                {
                    **action_decoders,
                    "fin_timeout": nested_kv_decoder(
                        KVDecoders(
                            {
                                "idle_timeout": decode_time,
                                "hard_timeout": decode_time,
                            }
                        )
                    ),
                }
            ),
            "clone": functools.partial(decode_exec, KVDecoders(action_decoders)),
        }

    @classmethod
    def _other_action_decoders(cls):
        """Recoders for other actions (see man(7) ovs-actions)"""
        return {
            "conjunction": nested_list_decoder(
                ListDecoders(
                    [("id", decode_int), ("k", decode_int), ("n", decode_int)]
                ),
                delims=[",", "/"],
            ),
            "note": decode_default,
            "sample": nested_kv_decoder(
                KVDecoders(
                    {
                        "probability": decode_int,
                        "collector_set_id": decode_int,
                        "obs_domain_id": decode_int,
                        "obs_point_id": decode_int,
                        "sampling_port": decode_default,
                        "ingress": decode_flag,
                        "egress": decode_flag,
                    }
                )
            ),
        }

    @classmethod
    def _act_decoders(cls):
        """Generate the actions decoders"""

        actions = {
            **cls._output_actions_decoders(),
            **cls._encap_actions_decoders(),
            **cls._field_action_decoders(),
            **cls._meta_action_decoders(),
            **cls._fw_action_decoders(),
            **cls._control_action_decoders(),
            **cls._other_action_decoders(),
        }
        clone_actions = cls._clone_actions_decoders(actions)
        actions.update(clone_actions)
        return KVDecoders(actions, default_free=decode_free_output)

    def __str__(self):
        if self._orig:
            return self._orig
        else:
            return self.to_string()

    def to_string(self):
        """Print a text representation of the flow"""
        string = "Info: {}\n" + self.info
        string += "Match : {}\n" + self.match
        string += "Actions: {}\n " + self.actions
        return string



import sys
import click
import colorsys
from rich.tree import Tree
from rich.text import Text
from rich.console import Console
from rich.style import Style
from rich.color import Color

from ovs_dbg.ofp import OFPFlow
from ovs_dbg.ofparse.main import maincli
from ovs_dbg.ofparse.process import process_flows, tojson, pprint
from ovs_dbg.ofparse.console import (
    ConsoleBuffer,
    ConsoleFormatter,
    hash_pallete,
    print_context,
)
from ovs_dbg.ofparse.html import HTMLBuffer, HTMLFormatter


@maincli.group(subcommand_metavar="FORMAT")
@click.pass_obj
def openflow(opts):
    """Process OpenFlow Flows"""
    pass


@openflow.command()
@click.pass_obj
def json(opts):
    """Print the flows in JSON format"""
    return tojson(flow_factory=create_ofp_flow, opts=opts)


@openflow.command()
@click.pass_obj
def pretty(opts):
    """Print the flows with some style"""
    return pprint(flow_factory=create_ofp_flow, opts=opts)


@openflow.command()
@click.option(
    "-s",
    "--show-flows",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show the full flows under each logical flow",
)
@click.pass_obj
def logic(opts, show_flows):
    """
    Print the logical structure of the flows.

    First, sorts the flows based on tables and priorities.
    Then, deduplicates logically equivalent flows: these a flows that match
    on the same set of fields (regardless of the values they match against),
    have the same priority, cookie and actions (regardless of action arguments)

    Frisorting the flows based on tables and priority, deduplicates flows
    based on
    """
    tables = dict()

    class LFlow:
        """A Logical Flow represents the scheleton of a flow

        Attributes:
            cookie (int): The flow cookie
            priority (int): The flow priority
            action_keys (tuple): The action keys
            match_keys (tuple): The match keys
        """

        def __init__(self, flow):
            self.priority = flow.match.get("priority") or 0
            self.cookie = flow.info.get("cookie") or 0
            self.action_keys = tuple([kv.key for kv in flow.actions_kv])
            self.match_keys = tuple([kv.key for kv in flow.match_kv])

        def __eq__(self, other):
            return (
                self.cookie == other.cookie
                and self.priority == other.priority
                and self.action_keys == other.action_keys
                and self.match_keys == other.match_keys
            )

        def __hash__(self):
            return tuple(
                [self.cookie, self.priority, self.action_keys, self.match_keys]
            ).__hash__()

    def callback(flow):
        """Parse the flows and sort them by table and logical flow"""
        table = flow.info.get("table") or 0
        if not tables.get(table):
            tables[table] = dict()

        # Group flows by logical hash
        lflow = LFlow(flow)

        if not tables[table].get(lflow):
            tables[table][lflow] = list()

        tables[table][lflow].append(flow)

    process_flows(
        flow_factory=create_ofp_flow,
        callback=callback,
        filename=opts.get("filename"),
        filter=opts.get("filter"),
    )

    # Try to make it easy to spot same cookies by printing them in different
    # colors
    cookie_style_gen = hash_pallete(
        hue=[x / 10 for x in range(0, 10)],
        saturation=[0.5],
        value=[0.5 + x / 10 * (0.85 - 0.5) for x in range(0, 10)],
    )

    tree = Tree("Ofproto Flows (logical)")
    console = Console(color_system=None if opts["style"] is None else "256")

    for table_num in sorted(tables.keys()):
        table = tables[table_num]
        table_tree = tree.add("** TABLE {} **".format(table_num))

        for lflow in sorted(
            table.keys(),
            key=(lambda x: x.priority),
            reverse=True,
        ):
            flows = table[lflow]

            buf = ConsoleBuffer(Text())

            buf.append_extra(
                "cookie={} ".format(hex(lflow.cookie)).ljust(18),
                style=cookie_style_gen(str(lflow.cookie)),
            )
            buf.append_extra("priority={} ".format(lflow.priority), style="steel_blue")
            buf.append_extra(",".join(lflow.match_keys), style="steel_blue")
            buf.append_extra("  --->  ", style="bold magenta")
            buf.append_extra(",".join(lflow.action_keys), style="steel_blue")
            buf.append_extra(" ( x {} )".format(len(flows)), style="dark_olive_green3")
            lflow_tree = table_tree.add(buf.text)

            if show_flows:
                for flow in flows:
                    buf = ConsoleBuffer(Text())
                    highlighted = None
                    if opts.get("highlight"):
                        result = opts.get("highlight").evaluate(flow)
                        if result:
                            highlighted = result.kv
                    ConsoleFormatter(opts=opts, console=console).format_flow(buf, flow, highlighted)
                    lflow_tree.add(buf.text)

    with print_context(console, opts):
        console.print(tree)


@openflow.command()
@click.pass_obj
def html(opts):
    """Print the flows in an HTML list"""
    tables = dict()

    def callback(flow):
        """Parse the flows and sort them by table"""
        table = flow.info.get("table") or 0
        if not tables.get(table):
            tables[table] = list()
        tables[table].append(flow)

    process_flows(
        flow_factory=create_ofp_flow,
        callback=callback,
        filename=opts.get("filename"),
        filter=opts.get("filter"),
    )

    html_obj = "<div id=flow_list>"
    for table, flows in tables.items():
        html_obj += "<h2 id=table_{table}> Table {table}</h2>".format(table=table)
        html_obj += "<ul id=table_{}_flow_list>".format(table)
        for flow in flows:
            html_obj += "<li id=flow_{}>".format(flow.id)
            highlighted = None
            if opts.get("highlight"):
                result = opts.get("highlight").evaluate(flow)
                if result:
                    highlighted = result.kv
            buf = HTMLBuffer()
            HTMLFormatter(opts).format_flow(buf, flow, highlighted)
            html_obj += buf.text
            html_obj += "</li>"
        html_obj += "</ul>"
    html_obj += "</div>"
    print(html_obj)


def create_ofp_flow(string, idx):
    """Create a OFPFlow"""
    if " reply " in string:
        return None
    return OFPFlow.from_string(string, idx)



""" Defines a Flow Filtering syntax
"""
import pyparsing as pp
import netaddr
from functools import reduce
from operator import and_, or_

from ovs_dbg.decoders import decode_default, decode_int, Decoder, IPMask, EthMask

from ovs_dbg.fields import field_decoders


class EvaluationResult:
    """An EvaluationResult is the result of an evaluation. It contains the
    boolean result and the list of key-values that were evaluated

    Note that since boolean operations (and, not, or) are based only on __bool__
    we use bitwise alternatives (&, ||, ~)
    """

    def __init__(self, result, *kv):
        self.result = result
        self.kv = kv if kv else list()

    def __and__(self, other):
        """Logical and operation"""
        return EvaluationResult(self.result and other.result, *self.kv, *other.kv)

    def __or__(self, other):
        """Logical or operation"""
        return EvaluationResult(self.result or other.result, *self.kv, *other.kv)

    def __invert__(self):
        """Logical not operation"""
        return EvaluationResult(not self.result, *self.kv)

    def __bool__(self):
        """Boolean operation"""
        return self.result

    def __repr__(self):
        return "{} [{}]".format(self.result, self.kv)


class ClauseExpression:
    operators = {}
    type_decoders = {
        int: decode_int,
        netaddr.IPAddress: IPMask,
        netaddr.EUI: EthMask,
        bool: bool,
    }

    def __init__(self, tokens):
        self.field = tokens[0]
        self.value = ""
        self.operator = ""

        if len(tokens) > 1:
            self.operator = tokens[1]
            self.value = tokens[2]

    def __repr__(self):
        return "{}(field: {}, operator: {}, value: {})".format(
            self.__class__.__name__, self.field, self.operator, self.value
        )

    def _find_data_in_kv(self, kv_list):
        """Find a KeyValue for evaluation in a list of KeyValue

        Args:
            kv_list (list[KeyValue]): list of KeyValue to look into
        """
        key_parts = self.field.split(".")
        field = key_parts[0]
        kvs = [kv for kv in kv_list if kv.key == field]
        if not kvs:
            return None

        for kv in kvs:
            if kv.key == self.field:
                # exact match
                return kv
            if len(key_parts) > 1:
                data = kv.value
                for subkey in key_parts[1:]:
                    try:
                        data = data.get(subkey)
                    except Exception:
                        data = None
                        break
                    if not data:
                        break
                if data:
                    return kv
        return None

    def _find_keyval_to_evaluate(self, flow):
        """Finds the key-value to use for evaluation"""
        for section in flow.sections:
            data = self._find_data_in_kv(section.data)
            if data:
                return data
        return None

    def evaluate(self, flow):
        """
        Return whether the clause is satisfied by the flow

        Args:
            flow (Flow): the flow to evaluate
        """
        keyval = self._find_keyval_to_evaluate(flow)

        if not keyval:
            return EvaluationResult(False)

        data = keyval.value

        if not self.value and not self.operator:
            # just asserting the existance of the key
            return EvaluationResult(True, keyval)

        # Decode the value based on the type of data
        if isinstance(data, Decoder):
            decoder = data.__class__
        else:
            decoder = self.type_decoders.get(data.__class__) or decode_default

        decoded_value = decoder(self.value)

        if self.operator == "=":
            return EvaluationResult(decoded_value == data, keyval)
        elif self.operator == "<":
            return EvaluationResult(data < decoded_value, keyval)
        elif self.operator == ">":
            return EvaluationResult(data > decoded_value, keyval)
        elif self.operator == "~=":
            return EvaluationResult(decoded_value in data, keyval)


class BoolNot:
    def __init__(self, t):
        self.op, self.args = t[0]

    def __repr__(self):
        return "NOT({})".format(self.args)

    def evaluate(self, flow):
        return ~self.args.evaluate(flow)


class BoolAnd:
    def __init__(self, pattern):
        self.args = pattern[0][0::2]

    def __repr__(self):
        return "AND({})".format(self.args)

    def evaluate(self, flow):
        return reduce(and_, [arg.evaluate(flow) for arg in self.args])


class BoolOr:
    def __init__(self, pattern):
        self.args = pattern[0][0::2]

    def evaluate(self, flow):
        return reduce(or_, [arg.evaluate(flow) for arg in self.args])

    def __repr__(self):
        return "OR({})".format(self.args)


class OFFilter:
    w = pp.Word(pp.alphanums + "." + ":" + "_" + "/" + "-")
    operators = (
        pp.Literal("=")
        | pp.Literal("~=")
        | pp.Literal("<")
        | pp.Literal(">")
        | pp.Literal("!=")
    )

    clause = (w + operators + w) | w
    clause.setParseAction(ClauseExpression)

    statement = pp.infixNotation(
        clause,
        [
            ("!", 1, pp.opAssoc.RIGHT, BoolNot),
            ("not", 1, pp.opAssoc.RIGHT, BoolNot),
            ("&&", 2, pp.opAssoc.LEFT, BoolAnd),
            ("and", 2, pp.opAssoc.LEFT, BoolAnd),
            ("||", 2, pp.opAssoc.LEFT, BoolOr),
            ("or", 2, pp.opAssoc.LEFT, BoolOr),
        ],
    )

    def __init__(self, expr):
        self._filter = self.statement.parseString(expr)

    def evaluate(self, flow):
        return self._filter[0].evaluate(flow)



""" Defines an Openvswitch Datapath Flow
"""
from functools import partial
from dataclasses import dataclass

from ovs_dbg.flow import Flow, Section

from ovs_dbg.kv import (
    KVParser,
    KVDecoders,
    ParseError,
    nested_kv_decoder,
    decode_nested_kv,
)
from ovs_dbg.decoders import (
    decode_default,
    decode_time,
    decode_int,
    decode_mask,
    Mask8,
    Mask16,
    Mask32,
    Mask128,
    IPMask,
    EthMask,
    decode_free_output,
    decode_flag,
    decode_ip_port_range,
    decode_nat,
)


class ODPFlow(Flow):
    """Datapath Flow"""

    def __init__(self, sections, raw="", id=None):
        """Constructor"""
        super(ODPFlow, self).__init__(sections, raw, id)

    @classmethod
    def from_string(cls, odp_string, id=None):
        """Parse a odp flow string

        The string is expected to have the follwoing format:
             [ufid], [match] [flow data] actions:[actions]

        Args:
            odp_string (str): a datapath flow string

        Returns:
            an ODPFlow instance
        """

        sections = []

        # If UFID present, parse it and
        ufid_pos = odp_string.find("ufid:")
        if ufid_pos >= 0:
            ufid_string = odp_string[ufid_pos : (odp_string[ufid_pos:].find(",") + 1)]
            ufid_parser = KVParser(KVDecoders({"ufid": decode_default}))
            ufid_parser.parse(ufid_string)
            if len(ufid_parser.kv()) != 1:
                raise ValueError("malformed odp flow: {}", odp_string)
            sections.append(Section("ufid", ufid_pos, ufid_string, ufid_parser.kv()))

        action_pos = odp_string.find("actions:")
        if action_pos < 0:
            raise ValueError("malformed odp flow: {}", odp_string)

        # rest of the string is between ufid and actions
        rest = odp_string[
            (ufid_pos + len(ufid_string) if ufid_pos >= 0 else 0) : action_pos
        ]

        action_pos += 8  # len("actions:")
        actions = odp_string[action_pos:]

        field_parts = rest.lstrip(" ").partition(" ")

        if len(field_parts) != 3:
            raise ValueError("malformed odp flow: {}", odp_string)

        match = field_parts[0]
        info = field_parts[2]

        info_decoders = cls._info_decoders()
        iparser = KVParser(KVDecoders(info_decoders))
        iparser.parse(info)
        isection = Section(
            name="info", pos=odp_string.find(info), string=info, data=iparser.kv()
        )
        sections.append(isection)

        mparser = cls._match_parser()
        mparser.parse(match)
        msection = Section(
            name="match", pos=odp_string.find(match), string=match, data=mparser.kv()
        )
        sections.append(msection)

        aparser = cls._action_parser()
        aparser.parse(actions)
        asection = Section(
            name="actions",
            pos=action_pos,
            string=actions,
            data=aparser.kv(),
            is_list=True,
        )
        sections.append(asection)

        return cls(sections, odp_string, id)

    @classmethod
    def _action_parser(cls):
        _decoders = {
            "drop": decode_flag,
            "lb_output": decode_int,
            "trunc": decode_int,
            "recirc": decode_int,
            "userspace": nested_kv_decoder(
                KVDecoders(
                    {
                        "pid": decode_int,
                        "sFlow": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "vid": decode_int,
                                    "pcp": decode_int,
                                    "output": decode_int,
                                }
                            )
                        ),
                        "slow_path": decode_default,
                        "flow_sample": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "probability": decode_int,
                                    "collector_sed_id": decode_int,
                                    "obs_domain_id": decode_int,
                                    "obs_point_id": decode_int,
                                    "output_port": decode_int,
                                    "ingress": decode_flag,
                                    "egress": decode_flag,
                                }
                            )
                        ),
                        "ipfix": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "output_port": decode_int,
                                }
                            )
                        ),
                        "controller": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "reason": decode_int,
                                    "dont_send": decode_int,
                                    "continuation": decode_int,
                                    "recirc_id": decode_int,
                                    "rule_cookie": decode_int,
                                    "controller_id": decode_int,
                                    "max_len": decode_int,
                                }
                            )
                        ),
                        "userdata": decode_default,
                        "actions": decode_flag,
                        "tunnel_out_port": decode_int,
                        "push_eth": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "src": EthMask,
                                    "dst": EthMask,
                                    "type": decode_int,
                                }
                            )
                        ),
                        "pop_eth": decode_flag,
                    }
                )
            ),
            "set": nested_kv_decoder(KVDecoders(cls._field_decoders())),
            "push_vlan": nested_kv_decoder(
                KVDecoders(
                    {
                        "vid": decode_int,
                        "pcp": decode_int,
                        "cfi": decode_int,
                        "tpid": decode_int,
                    }
                )
            ),
            "pop_vlan": decode_flag,
            "push_nsh": nested_kv_decoder(
                KVDecoders(
                    {
                        "flags": decode_int,
                        "ttl": decode_int,
                        "mdtype": decode_int,
                        "np": decode_int,
                        "spi": decode_int,
                        "si": decode_int,
                        "c1": decode_int,
                        "c2": decode_int,
                        "c3": decode_int,
                        "c4": decode_int,
                        "md2": decode_int,
                    }
                )
            ),
            "pop_nsh": decode_flag,
            "tnl_pop": decode_int,
            "ct_clear": decode_flag,
            "ct": nested_kv_decoder(
                KVDecoders(
                    {
                        "commit": decode_flag,
                        "force_commit": decode_flag,
                        "zone": decode_int,
                        "mark": Mask32,
                        "label": Mask128,
                        "helper": decode_default,
                        "timeout": decode_default,
                        "nat": decode_nat,
                    }
                )
            ),
            **cls._tnl_action_decoder(),
        }

        _decoders["clone"] = nested_kv_decoder(
            KVDecoders(decoders=_decoders, default_free=decode_free_output)
        )

        return KVParser(
            KVDecoders(
                decoders={
                    **_decoders,
                    # "clone": nested_kv_decoder(KVDecoders(_decoders)),
                    "sample": nested_kv_decoder(
                        KVDecoders(
                            {
                                "sample": (lambda x: decode_int(x.strip("%"))),
                                "actions": nested_kv_decoder(
                                    KVDecoders(
                                        decoders=_decoders,
                                        default_free=decode_free_output,
                                    )
                                ),
                            }
                        )
                    ),
                    "check_pkt_len": nested_kv_decoder(
                        KVDecoders(
                            {
                                "size": decode_int,
                                "gt": nested_kv_decoder(
                                    KVDecoders(
                                        decoders=_decoders,
                                        default_free=decode_free_output,
                                    )
                                ),
                                "le": nested_kv_decoder(
                                    KVDecoders(
                                        decoders=_decoders,
                                        default_free=decode_free_output,
                                    )
                                ),
                            }
                        )
                    ),
                },
                default_free=decode_free_output,
            )
        )

    @classmethod
    def _tnl_action_decoder(cls):
        return {
            "tnl_push": nested_kv_decoder(
                KVDecoders(
                    {
                        "tnl_port": decode_int,
                        "header": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "size": decode_int,
                                    "type": decode_int,
                                    "eth": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "src": EthMask,
                                                "dst": EthMask,
                                                "dl_type": decode_int,
                                            }
                                        )
                                    ),
                                    "ipv4": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "src": IPMask,
                                                "dst": IPMask,
                                                "proto": decode_int,
                                                "tos": decode_int,
                                                "ttl": decode_int,
                                                "frag": decode_int,
                                            }
                                        )
                                    ),
                                    "ipv6": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "src": IPMask,
                                                "dst": IPMask,
                                                "label": decode_int,
                                                "proto": decode_int,
                                                "tclass": decode_int,
                                                "hlimit": decode_int,
                                            }
                                        )
                                    ),
                                    "udp": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "src": decode_int,
                                                "dst": decode_int,
                                                "dsum": Mask16,
                                            }
                                        )
                                    ),
                                    "vxlan": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "flags": decode_int,
                                                "vni": decode_int,
                                            }
                                        )
                                    ),
                                    "geneve": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "oam": decode_flag,
                                                "crit": decode_flag,
                                                "vni": decode_int,
                                                "options": partial(
                                                    decode_geneve, False
                                                ),
                                            }
                                        )
                                    ),
                                    "gre": decode_tnl_gre,
                                    "erspan": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "ver": decode_int,
                                                "sid": decode_int,
                                                "idx": decode_int,
                                                "sid": decode_int,
                                                "dir": decode_int,
                                                "hwid": decode_int,
                                            }
                                        )
                                    ),
                                    "gtpu": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "flags": decode_int,
                                                "msgtype": decode_int,
                                                "teid": decode_int,
                                            }
                                        )
                                    ),
                                }
                            )
                        ),
                        "out_port": decode_int,
                    }
                )
            )
        }

    @classmethod
    def _info_decoders(cls):
        return {
            "packets": decode_int,
            "bytes": decode_int,
            "used": decode_time,
            "flags": decode_default,
            "dp": decode_default,
        }

    @classmethod
    def _match_parser(cls):
        return KVParser(
            KVDecoders(
                {
                    **cls._field_decoders(),
                    "encap": nested_kv_decoder(KVDecoders(cls._field_decoders())),
                }
            )
        )

    @classmethod
    def _field_decoders(cls):
        return {
            "skb_priority": Mask32,
            "skb_mark": Mask32,
            "recirc_id": decode_int,
            "dp_hash": Mask32,
            "ct_state": decode_default,  # TODO: Parse flags
            "ct_zone": Mask16,
            "ct_mark": Mask32,
            "ct_label": Mask128,
            "ct_tuple4": nested_kv_decoder(
                KVDecoders(
                    {
                        "src": IPMask,
                        "dst": IPMask,
                        "proto": Mask8,
                        "tcp_src": Mask16,
                        "tcp_dst": Mask16,
                    }
                )
            ),
            "ct_tuple6": nested_kv_decoder(
                KVDecoders(
                    {
                        "src": IPMask,
                        "dst": IPMask,
                        "proto": Mask8,
                        "tcp_src": Mask16,
                        "tcp_dst": Mask16,
                    }
                )
            ),
            "tunnel": nested_kv_decoder(
                KVDecoders(
                    {
                        "tun_id": decode_int,
                        "src": IPMask,
                        "dst": IPMask,
                        "ipv6_src": IPMask,
                        "ipv6_dst": IPMask,
                        "tos": decode_int,
                        "ttl": decode_int,
                        "tp_src": decode_int,
                        "tp_dst": decode_int,
                        "erspan": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "ver": decode_int,
                                    "idx": decode_int,
                                    "sid": decode_int,
                                    "dir": decode_int,
                                    "hwid": decode_int,
                                }
                            )
                        ),
                        "vxlan": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "gbp": nested_kv_decoder(
                                        KVDecoders(
                                            {
                                                "id": decode_int,
                                                "flags": decode_int,
                                            }
                                        )
                                    )
                                }
                            )
                        ),
                        "geneve": partial(decode_geneve, True),
                        "gtpu": nested_kv_decoder(
                            KVDecoders(
                                {
                                    "flags": Mask8,
                                    "msgtype": Mask8,
                                }
                            )
                        ),
                        "flags": decode_default,
                    }
                )
            ),
            "in_port": decode_default,
            "eth": nested_kv_decoder(
                KVDecoders(
                    {
                        "src": EthMask,
                        "dst": EthMask,
                    }
                )
            ),
            "vlan": nested_kv_decoder(
                KVDecoders(
                    {
                        "vid": Mask16,
                        "pcp": Mask16,
                        "cfi": Mask16,
                    }
                )
            ),
            "eth_type": Mask16,
            "mpls": nested_kv_decoder(
                KVDecoders(
                    {
                        "label": Mask32,
                        "tc": Mask32,
                        "ttl": Mask32,
                        "bos": Mask32,
                    }
                )
            ),
            "ipv4": nested_kv_decoder(
                KVDecoders(
                    {
                        "src": IPMask,
                        "dst": IPMask,
                        "proto": Mask8,
                        "tos": Mask8,
                        "ttl": Mask8,
                        "frag": decode_default,
                    }
                )
            ),
            "ipv6": nested_kv_decoder(
                KVDecoders(
                    {
                        "src": IPMask,
                        "dst": IPMask,
                        "label": decode_mask(20),
                        "proto": Mask8,
                        "tclass": Mask8,
                        "hlimit": Mask8,
                        "frag": decode_default,
                    }
                )
            ),
            "tcp": nested_kv_decoder(
                KVDecoders(
                    {
                        "src": Mask16,
                        "dst": Mask16,
                    }
                )
            ),
            "tcp_flags": decode_default,
            "udp": nested_kv_decoder(
                KVDecoders(
                    {
                        "src": Mask16,
                        "dst": Mask16,
                    }
                )
            ),
            "sctp": nested_kv_decoder(
                KVDecoders(
                    {
                        "src": Mask16,
                        "dst": Mask16,
                    }
                )
            ),
            "icmp": nested_kv_decoder(
                KVDecoders(
                    {
                        "type": Mask8,
                        "code": Mask8,
                    }
                )
            ),
            "icmpv6": nested_kv_decoder(
                KVDecoders(
                    {
                        "type": Mask8,
                        "code": Mask8,
                    }
                )
            ),
            "arp": nested_kv_decoder(
                KVDecoders(
                    {
                        "sip": IPMask,
                        "tip": IPMask,
                        "op": Mask16,
                        "sha": EthMask,
                        "tha": EthMask,
                    }
                )
            ),
            "nd": nested_kv_decoder(
                KVDecoders(
                    {
                        "target": IPMask,
                        "sll": EthMask,
                        "tll": EthMask,
                    }
                )
            ),
            "nd_ext": nested_kv_decoder(
                KVDecoders(
                    {
                        "nd_reserved": Mask32,
                        "nd_options_type": Mask8,
                    }
                )
            ),
            "packet_type": nested_kv_decoder(
                KVDecoders(
                    {
                        "ns": Mask16,
                        "id": Mask16,
                    }
                )
            ),
            "nsh": nested_kv_decoder(
                KVDecoders(
                    {
                        "flags": Mask8,
                        "mdtype": Mask8,
                        "np": Mask8,
                        "spi": Mask32,
                        "si": Mask8,
                        "c1": Mask32,
                        "c2": Mask32,
                        "c3": Mask32,
                        "c4": Mask32,
                    }
                )
            ),
        }


def decode_geneve(mask, value):
    """
    Decode geneve options. Used for both tnl_push(header(geneve(options()))) action
    and tunnel(geneve()) match.

    It has the following format:

    {class=0xffff,type=0x80,len=4,0xa}

    Args:
        mask (bool): Whether masking is supported
        value (str): The value to decode
    """
    if mask:
        decoders = {
            "class": Mask16,
            "type": Mask8,
            "len": Mask8,
        }

        def free_decoder(value):
            return "data", Mask128(value)

    else:
        decoders = {
            "class": decode_int,
            "type": decode_int,
            "len": decode_int,
        }

        def free_decoder(value):
            return "data", decode_int(value)

    return decode_nested_kv(
        KVDecoders(decoders=decoders, default_free=free_decoder), value.strip("{}")
    )


def decode_tnl_gre(value):
    """
    Decode tnl_push(header(gre())) action

    It has the following format:

    gre((flags=0x2000,proto=0x6558),key=0x1e241))

    Args:
        value (str): The value to decode
    """

    return decode_nested_kv(
        KVDecoders(
            {
                "flags": decode_int,
                "proto": decode_int,
                "key": decode_int,
                "csum": decode_int,
                "seq": decode_int,
            }
        ),
        value.replace("(", "").replace(")", ""),
    )



""" This module defines console formatting
"""

import colorsys
import contextlib
import itertools
import sys
import zlib
from rich.console import Console
from rich.text import Text
from rich.style import Style
from rich.color import Color

from ovs_dbg.ofparse.format import FlowFormatter, FlowBuffer, FlowStyle


class ConsoleBuffer(FlowBuffer):
    """ConsoleBuffer implements FlowBuffer to provide console-based text
    formatting based on rich.Text

    Append functions accept a rich.Style

    Args:
        rtext(rich.Text): Optional; text instance to reuse
    """

    def __init__(self, rtext):
        self._text = rtext or Text()

    @property
    def text(self):
        return self._text

    def _append(self, string, style):
        """Append to internal text"""
        return self._text.append(string, style)

    def append_key(self, kv, style):
        """Append a key
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (rich.Style): the style to use
        """
        return self._append(kv.meta.kstring, style)

    def append_delim(self, kv, style):
        """Append a delimiter
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (rich.Style): the style to use
        """
        return self._append(kv.meta.delim, style)

    def append_end_delim(self, kv, style):
        """Append an end delimiter
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (rich.Style): the style to use
        """
        return self._append(kv.meta.end_delim, style)

    def append_value(self, kv, style):
        """Append a value
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (rich.Style): the style to use
        """
        return self._append(kv.meta.vstring, style)

    def append_extra(self, extra, style):
        """Append extra string
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (rich.Style): the style to use
        """
        return self._append(extra, style)


class ConsoleFormatter(FlowFormatter):
    """
    Args:
        console (rich.Console): Optional, an existing console to use
        max_value_len (int): Optional; max length of the printed values
        kwargs (dict): Optional; Extra arguments to be passed down to
            rich.console.Console()
    """

    def __init__(self, opts=None, console=None, **kwargs):
        super(ConsoleFormatter, self).__init__()
        style = self.style_from_opts(opts)
        self.console = console or Console(no_color=(style is None), **kwargs)
        self.style = style or FlowStyle()

    def style_from_opts(self, opts):
        return self._style_from_opts(opts, "console", Style)

    def print_flow(self, flow, highlighted=None):
        """
        Prints a flow to the console

        Args:
            flow (ovs_dbg.OFPFlow): the flow to print
            style (dict): Optional; style dictionary to use
            highlighted (list): Optional; list of KeyValues to highlight
        """

        buf = ConsoleBuffer(Text())
        self.format_flow(buf, flow, highlighted)
        self.console.print(buf.text)

    def format_flow(self, buf, flow, highlighted=None):
        """
        Formats the flow into the provided buffer as a rich.Text

        Args:
            buf (FlowBuffer): the flow buffer to append to
            flow (ovs_dbg.OFPFlow): the flow to format
            style (FlowStyle): Optional; style object to use
            highlighted (list): Optional; list of KeyValues to highlight
        """
        return super(ConsoleFormatter, self).format_flow(
            buf, flow, self.style, highlighted
        )


def hash_pallete(hue, saturation, value):
    """Generates a color pallete with the cartesian product
    of the hsv values provided and returns a callable that assigns a color for
    each value hash
    """
    HSV_tuples = itertools.product(hue, saturation, value)
    RGB_tuples = map(lambda x: colorsys.hsv_to_rgb(*x), HSV_tuples)
    styles = [
        Style(color=Color.from_rgb(r * 255, g * 255, b * 255)) for r, g, b in RGB_tuples
    ]

    def get_style(string):
        hash_val = zlib.crc32(bytes(str(string), "utf-8"))
        return styles[hash_val % len(styles)]

    return get_style


def print_context(console, opts):
    """
    Returns a printing context

    Args:
        console: The console to print
        paged (bool): Wheter to page the output
        style (bool): Whether to force the use of styled pager
    """
    if opts.get("paged"):
        # Internally pydoc's pager library is used which returns a
        # plain pager if both stdin and stdout are not tty devices
        #
        # Workaround that limitation if only stdin is not a tty (e.g
        # data is piped to us through stdin)
        if not sys.stdin.isatty() and sys.stdout.isatty():
            setattr(sys.stdin, "isatty", lambda: True)

        with_style = opts.get("style") is not None

        return console.pager(styles=with_style)

    return contextlib.nullcontext()



""" Defines decoders for openflow actions
"""

import netaddr
import functools

from ovs_dbg.kv import nested_kv_decoder, KVDecoders, KeyValue, KVParser
from ovs_dbg.decoders import decode_default, decode_time, decode_flag, decode_int
from ovs_dbg.fields import field_decoders


def decode_output(value):
    """Decodes the output value

    Does not support field specification
    """
    if len(value.split(",")) > 1:
        return nested_kv_decoder()(value)
    try:
        return {"port": int(value)}
    except ValueError:
        return {"port": value.strip('"')}


def decode_controller(value):
    """Decodes the controller action"""
    if not value:
        return KeyValue("output", "controller")
    else:
        # Try controller:max_len
        try:
            max_len = int(value)
            return {
                "max_len": max_len,
            }
        except ValueError:
            pass
        # controller(key[=val], ...)
        return nested_kv_decoder()(value)


def decode_bundle_load(value):
    return decode_bundle(value, True)


def decode_bundle(value, load=False):
    """Decode bundle action"""
    result = {}
    keys = ["fields", "basis", "algorithm", "ofport"]
    if load:
        keys.append("dst")

    for key in keys:
        parts = value.partition(",")
        nvalue = parts[0]
        value = parts[2]
        if key == "ofport":
            continue
        result[key] = decode_default(nvalue)

    # Handle members:
    mvalues = value.split("members:")
    result["members"] = [int(port) for port in mvalues[1].split(",")]
    return result


def decode_encap_ethernet(value):
    """Decodes encap ethernet value"""
    return "ethernet", int(value, 0)


def decode_field(value):
    """Decodes a field as defined in the 'Field Specification' of the actions
    man page: http://www.openvswitch.org/support/dist-docs/ovs-actions.7.txt
    """
    parts = value.strip("]\n\r").split("[")
    result = {
        "field": parts[0],
    }

    if len(parts) > 1 and parts[1]:
        field_range = parts[1].split("..")
        start = field_range[0]
        end = field_range[1] if len(field_range) > 1 else start
        if start:
            result["start"] = int(start)
        if end:
            result["end"] = int(end)

    return result


def decode_load_field(value):
    """Decodes 'load:value->dst' actions"""
    parts = value.split("->")
    if len(parts) != 2:
        raise ValueError("Malformed load action : %s" % value)

    return {"value": int(parts[0], 0), "dst": decode_field(parts[1])}


def decode_set_field(field_decoders, value):
    """Decodes 'set_field:value/mask->dst' actions

    The value is decoded by field_decoders which is a KVDecoders instance
    Args:
        field_decoders
    """
    parts = value.split("->")
    if len(parts) != 2:
        raise ValueError("Malformed set_field action : %s" % value)

    val = parts[0]
    dst = parts[1]

    val_result = field_decoders.decode(dst, val)

    return {
        "value": {val_result[0]: val_result[1]},
        "dst": decode_field(dst),
    }


def decode_move_field(value):
    """Decodes 'move:src->dst' actions"""
    parts = value.split("->")
    if len(parts) != 2:
        raise ValueError("Malformed move action : %s" % value)

    return {
        "src": decode_field(parts[0]),
        "dst": decode_field(parts[1]),
    }


def decode_dec_ttl(value):
    """Decodes dec_ttl and dec_ttl(id, id[2], ...) actions"""
    if not value:
        return True
    return [int(idx) for idx in value.split(",")]


def decode_chk_pkt_larger(value):
    """Decodes 'check_pkt_larger(pkt_len)->dst' actions"""
    parts = value.split("->")
    if len(parts) != 2:
        raise ValueError("Malformed check_pkt_larger action : %s" % value)

    pkt_len = int(parts[0].strip("()"))
    dst = decode_field(parts[1])
    return {"pkt_len": pkt_len, "dst": dst}


# CT decoders
def decode_zone(value):
    """Decodes the 'zone' keyword of the ct action"""
    try:
        return int(value, 0)
    except ValueError:
        pass
    return decode_field(value)


def decode_exec(action_decoders, value):
    """Decodes the 'exec' keyword of the ct action

    Args:
        decode_actions (KVDecoders): the decoders to be used to decode the
            nested exec
        value (string): the string to be decoded
    """
    exec_parser = KVParser(action_decoders)
    exec_parser.parse(value)
    return [{kv.key: kv.value} for kv in exec_parser.kv()]


def decode_learn(action_decoders):
    """Create the decoder to be used to decode the 'learn' action.

    The learn action can include any nested action, therefore we need decoders
    for all possible actions.

    Args:
        action_decoders (dict): dictionary of decoders to be used in nested
            action decoding

    """

    def decode_learn_field(decoder, value):
        """Generates a decoder to be used for the 'field' argument of the
        'learn' action.

        The field can hold a value that should be decoded, either as a field,
        or as a the value (see man(7) ovs-actions)

        Args:
            decoder (callable): The decoder

        """
        if value in field_decoders.keys():
            # It's a field
            return value
        else:
            return decoder(value)

    learn_field_decoders = {
        field: functools.partial(decode_learn_field, decoder)
        for field, decoder in field_decoders.items()
    }
    learn_decoders = {
        **action_decoders,
        **learn_field_decoders,
        "idle_timeout": decode_time,
        "hard_timeout": decode_time,
        "priority": decode_int,
        "cooke": decode_int,
        "send_flow_rem": decode_flag,
        "table": decode_int,
        "delete_learned": decode_flag,
        "limit": decode_int,
        "result_dst": decode_field,
    }

    return functools.partial(decode_exec, KVDecoders(learn_decoders))



""" Defines the Flow class
"""

from ovs_dbg.kv import KeyValue


class Section(object):
    """A section within a Flow

    Attributes:
        name (str): name of the section
        pos (int): position within the overall flow string
        string (str): section string
        data (list[KeyValue]): parsed data of the section
        is_list (bool): whether the key-values shall be expressed as a list
            (e.g: it allows repeated keys)
    """

    def __init__(self, name, pos, string, data, is_list=False):
        self.name = name
        self.pos = pos
        self.string = string
        self.data = data
        self.is_list = is_list

    def __str__(self):
        return "{} (at {}): {}".format(self.name, self.pos, self.string)

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self)

    def dict(self):
        return {self.name: self.format_data()}

    def format_data(self):
        if self.is_list:
            return [{item.key: item.value} for item in self.data]
        else:
            return {item.key: item.value for item in self.data}


class Flow(object):
    """The Flow class is a base class for other types of concrete flows
    (such as OFproto Flows or DPIF Flows)

    For each section provided, the object will have the following attributes
    {section} will return the sections data in a formatted way
    {section}_kv will return the sections data as a list of KeyValues

    Args:
        sections (list[Section]): list of sections that comprise the flow
        orig (str): Original flow string
        id (Any): Identifier
    """

    def __init__(self, sections, orig="", id=None):
        self._sections = sections
        self._orig = orig
        self._id = id
        for section in sections:
            setattr(self, section.name, self.section(section.name).format_data())
            setattr(self, "{}_kv".format(section.name), self.section(section.name).data)

    def section(self, name):
        """Return the section by name"""
        return next((sect for sect in self._sections if sect.name == name), None)

    @property
    def id(self):
        """Return the Flow ID"""
        return self._id

    @property
    def sections(self):
        """Return the section by name"""
        return self._sections

    @property
    def orig(self):
        return self._orig

    def dict(self):
        """Returns the Flow information in a dictionary"""
        flow_dict = {"orig": self.orig}
        for section in self.sections:
            flow_dict.update(section.dict())

        return flow_dict



""" This module defines generic formatting classes
"""


class FlowStyle:
    """
    A FlowStyle determines the KVStyle to use for each key value in a flow

    Styles are internally represented by a dictionary.
    In order to determine the style for a "key", the following items in the
    dictionary are fetched:
        - key.highlighted.{key} (if key is found in hightlighted)
        - key.highlighted (if key is found in hightlighted)
        - key.{key}
        - key
        - default

    In order to determine the style for a "value", the following items in the
    dictionary are fetched:
        - value.highlighted.{key} (if key is found in hightlighted)
        - value.highlighted.type{value.__class__.__name__}
        - value.highlighted
        (if key is found in hightlighted)
        - value.{key}
        - value.type.{value.__class__.__name__}
        - value
        - default

    Additionally, the following style items can be defined:
        - delim: for delimiters
        - delim.highlighted: for delimiters of highlighted key-values
    """

    def __init__(self, initial=None):
        self._styles = initial if initial is not None else dict()

    def set_flag_style(self, kvstyle):
        self._styles["flag"] = kvstyle

    def set_delim_style(self, kvstyle, highlighted=False):
        if highlighted:
            self._styles["delim.highlighted"] = kvstyle
        else:
            self._styles["delim"] = kvstyle

    def set_default_key_style(self, kvstyle):
        self._styles["key"] = kvstyle

    def set_default_value_style(self, kvstyle):
        self._styles["value"] = kvstyle

    def set_key_style(self, key, kvstyle, highlighted=False):
        if highlighted:
            self._styles["key.highlighted.{}".format(key)] = kvstyle
        else:
            self._styles["key.{}".format(key)] = kvstyle

    def set_value_style(self, key, kvstyle, highlighted=None):
        if highlighted:
            self._styles["value.highlighted.{}".format(key)] = kvstyle
        else:
            self._styles["value.{}".format(key)] = kvstyle

    def set_value_type_style(self, name, kvstyle, highlighted=None):
        if highlighted:
            self._styles["value.highlighted.type.{}".format(name)] = kvstyle
        else:
            self._styles["value.type.{}".format(name)] = kvstyle

    def get(self, key):
        return self._styles.get(key)

    def get_delim_style(self, highlighted=False):
        delim_style_lookup = ["delim.highlighted"] if highlighted else []
        delim_style_lookup.extend(["delim", "default"])
        return next(
            (self._styles.get(s) for s in delim_style_lookup if self._styles.get(s)),
            None,
        )

    def get_flag_style(self):
        return self._styles.get("flag") or self._styles.get("default")

    def get_key_style(self, kv, highlighted=False):
        key = kv.meta.kstring

        key_style_lookup = (
            ["key.highlighted.%s" % key, "key.highlighted"] if highlighted else []
        )
        key_style_lookup.extend(["key.%s" % key, "key", "default"])

        style = next(
            (self._styles.get(s) for s in key_style_lookup if self._styles.get(s)),
            None,
        )
        if callable(style):
            return style(kv.meta.kstring)
        return style

    def get_value_style(self, kv, highlighted=False):
        key = kv.meta.kstring
        value_type = kv.value.__class__.__name__
        value_style_lookup = (
            [
                "value.highlighted.%s" % key,
                "value.highlighted.type.%s" % value_type,
                "value.highlighted",
            ]
            if highlighted
            else []
        )
        value_style_lookup.extend(
            [
                "value.%s" % key,
                "value.type.%s" % value_type,
                "value",
                "default",
            ]
        )

        style = next(
            (self._styles.get(s) for s in value_style_lookup if self._styles.get(s)),
            None,
        )
        if callable(style):
            return style(kv.meta.vstring)
        return style


class FlowFormatter:
    """FlowFormatter is a base class for Flow Formatters"""

    def __init__(self):
        self._highlighted = list()

    def _style_from_opts(self, opts, opts_key, style_constructor):
        """Create style object from options

        Args:
            opts (dict): Options dictionary
            opts_key (str): The options style key to extract (e.g: console or html)
            style_constructor(callable): A callable that creates a derived style object
        """
        if not opts or not opts.get("style"):
            return None

        section_name = ".".join(["styles", opts.get("style")])
        if section_name not in opts.get("config").sections():
            raise Exception("Style not present in config file")

        config = opts.get("config")[section_name]
        style = {}
        for key in config:
            (_, console, style_full_key) = key.partition(opts_key + ".")
            if not console:
                continue

            (style_key, _, prop) = style_full_key.rpartition(".")
            if not prop or not style_key:
                raise Exception("malformed style config: {}".format(key))

            if not style.get(style_key):
                style[style_key] = {}
            style[style_key][prop] = config[key]

        return FlowStyle({k: style_constructor(**v) for k, v in style.items()})

    def format_flow(self, buf, flow, style_obj=None, highlighted=None):
        """
        Formats the flow into the provided buffer

        Args:
            buf (FlowBuffer): the flow buffer to append to
            flow (ovs_dbg.OFPFlow): the flow to format
            style_obj (FlowStyle): Optional; style to use
            highlighted (list): Optional; list of KeyValues to highlight
        """
        last_printed_pos = 0
        style_obj = style_obj or FlowStyle()

        for section in sorted(flow.sections, key=lambda x: x.pos):
            buf.append_extra(
                flow.orig[last_printed_pos : section.pos],
                style=style_obj.get("default"),
            )
            self.format_kv_list(
                buf, section.data, section.string, style_obj, highlighted
            )
            last_printed_pos = section.pos + len(section.string)

    def format_kv_list(self, buf, kv_list, full_str, style_obj, highlighted):
        """
        Format a KeyValue List

        Args:
            buf (FlowBuffer): a FlowBuffer to append formatted KeyValues to
            kv_list (list[KeyValue]: the KeyValue list to format
            full_str (str): the full string containing all k-v
            style_obj (FlowStyle): a FlowStyle object to use
            highlighted (list): Optional; list of KeyValues to highlight
        """
        for i in range(len(kv_list)):
            kv = kv_list[i]
            written = self.format_kv(
                buf, kv, style_obj=style_obj, highlighted=highlighted
            )

            end = kv_list[i + 1].meta.kpos if i < (len(kv_list) - 1) else len(full_str)

            buf.append_extra(
                full_str[(kv.meta.kpos + written) : end].rstrip("\n\r"),
                style=style_obj.get("default"),
            )

    def format_kv(self, buf, kv, style_obj, highlighted=None):
        """Format a KeyValue

        A formatted keyvalue has the following parts:
            {key}{delim}{value}[{delim}]

        Args:
            buf (FlowBuffer): buffer to append the KeyValue to
            kv (KeyValue): The KeyValue to print
            style_obj (FlowStyle): The style object to use
            highlighted (list): Optional; list of KeyValues to highlight

        Returns the number of printed characters
        """
        ret = 0
        key = kv.meta.kstring
        is_highlighted = key in [k.key for k in highlighted] if highlighted else False

        key_style = style_obj.get_key_style(kv, is_highlighted)
        buf.append_key(kv, key_style)  # format value
        ret += len(key)

        if not kv.meta.vstring:
            return ret

        if kv.meta.delim not in ("\n", "\t", "\r", ""):
            buf.append_delim(kv, style_obj.get_delim_style(is_highlighted))
            ret += len(kv.meta.delim)

        value_style = style_obj.get_value_style(kv, is_highlighted)

        buf.append_value(kv, value_style)  # format value
        ret += len(kv.meta.vstring)

        if kv.meta.end_delim:
            buf.append_end_delim(kv, style_obj.get_delim_style(is_highlighted))
            ret += len(kv.meta.end_delim)

        return ret


class FlowBuffer:
    """A FlowBuffer is a base class for format buffers.
    Childs must implement the following methods:
        append_key(self, kv, style)
        append_value(self, kv, style)
        append_delim(self, delim, style)
        append_end_delim(self, delim, style)
        append_extra(self, extra, style)
    """

    def append_key(self, kv, style):
        """Append a key
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise Exception("Not implemented")

    def append_delim(self, kv, style):
        """Append a delimiter
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise Exception("Not implemented")

    def append_end_delim(self, kv, style):
        """Append an end delimiter
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise Exception("Not implemented")

    def append_value(self, kv, style):
        """Append a value
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise Exception("Not implemented")

    def append_extra(self, extra, style):
        """Append extra string
        Args:
            kv (KeyValue): the KeyValue instance to append
            style (Any): the style to use
        """
        raise Exception("Not implemented")



import sys
import click
import colorsys
import graphviz
import itertools

from rich.tree import Tree
from rich.text import Text
from rich.console import Console
from rich.style import Style
from rich.color import Color

from ovs_dbg.ofparse.main import maincli
from ovs_dbg.ofparse.process import process_flows, tojson, pprint
from ovs_dbg.ofparse.console import (
    ConsoleFormatter,
    ConsoleBuffer,
    print_context,
    hash_pallete,
)
from ovs_dbg.ofparse.format import FlowStyle
from ovs_dbg.ofparse.html import HTMLBuffer, HTMLFormatter
from ovs_dbg.odp import ODPFlow
from ovs_dbg.filter import OFFilter


@maincli.group(subcommand_metavar="FORMAT")
@click.pass_obj
def datapath(opts):
    """Process DPIF Flows"""
    pass


@datapath.command()
@click.pass_obj
def json(opts):
    """Print the flows in JSON format"""
    return tojson(flow_factory=ODPFlow.from_string, opts=opts)


@datapath.command()
@click.pass_obj
def pretty(opts):
    """Print the flows with some style"""
    return pprint(flow_factory=ODPFlow.from_string, opts=opts)


@datapath.command()
@click.pass_obj
def logic(opts):
    """Print the flows in a tree based on the 'recirc_id'"""

    flow_list = []

    def callback(flow):
        flow_list.append(flow)

    process_flows(
        flow_factory=ODPFlow.from_string,
        callback=callback,
        filename=opts.get("filename"),
        filter=opts.get("filter"),
    )

    tree = Tree("Datapath Flows (logical)")
    ofconsole = ConsoleFormatter(opts)
    console = ofconsole.console

    # HSV_tuples = [(x / size, 0.7, 0.8) for x in range(size)]
    recirc_style_gen = hash_pallete(
        hue=[x / 50 for x in range(0, 50)], saturation=[0.7], value=[0.8]
    )

    style = ofconsole.style
    style.set_default_value_style(Style(color="bright_black"))
    style.set_key_style("output", Style(color="green"))
    style.set_value_style("output", Style(color="green"))
    style.set_value_style("recirc", recirc_style_gen)
    style.set_value_style("recirc_id", recirc_style_gen)

    def append_to_tree(parent, flow):
        buf = ConsoleBuffer(Text())
        highlighted = None
        if opts.get("highlight"):
            result = opts.get("highlight").evaluate(flow)
            if result:
                highlighted = result.kv
        ofconsole.format_flow(buf, flow, highlighted)
        tree_elem = parent.add(buf.text)
        return tree_elem

    process_flow_tree(flow_list, tree, 0, append_to_tree)

    with print_context(console, opts):
        console.print(tree)


@datapath.command()
@click.pass_obj
def html(opts):
    """Print the flows in an HTML list sorted by recirc_id"""

    flow_list = []

    def callback(flow):
        flow_list.append(flow)

    process_flows(
        flow_factory=ODPFlow.from_string,
        callback=callback,
        filename=opts.get("filename"),
        filter=opts.get("filter"),
    )

    html_obj = get_html_obj(flow_list, opts)

    print(html_obj)


@datapath.command()
@click.option(
    "-h",
    "--html",
    is_flag=True,
    default=False,
    show_default=True,
    help="Output an html file containing the graph",
)
@click.pass_obj
def graph(opts, html):
    """Print the flows in an graphviz (.dot) format showing the relationship of recirc_ids"""

    recirc_flows = {}

    def callback(flow):
        """Parse the flows and sort them by table"""
        rid = hex(flow.match.get("recirc_id") or 0)
        if not recirc_flows.get(rid):
            recirc_flows[rid] = list()
        recirc_flows[rid].append(flow)

    process_flows(
        flow_factory=ODPFlow.from_string,
        callback=callback,
        filename=opts.get("filename"),
        filter=opts.get("filter"),
    )

    node_styles = {
        OFFilter("ct and (ct_state or ct_label or ct_mark)"): {"color": "#ff00ff"},
        OFFilter("ct_state or ct_label or ct_mark"): {"color": "#0000ff"},
        OFFilter("ct"): {"color": "#ff0000"},
    }

    g = graphviz.Digraph("DP flows", node_attr={"shape": "rectangle"})
    g.attr(compound="true")
    g.attr(rankdir="TB")

    for recirc, flows in recirc_flows.items():
        with g.subgraph(
            name="cluster_{}".format(recirc), comment="recirc {}".format(recirc)
        ) as sg:

            sg.attr(rankdir="TB")
            sg.attr(ranksep="0.02")
            sg.attr(label="recirc_id {}".format(recirc))

            invis = "f{}".format(recirc)
            sg.node(invis, color="white", len="0", shape="point", width="0", height="0")

            previous = None
            for flow in flows:
                name = "Flow_{}".format(flow.id)
                summary = "Line: {} \n".format(flow.id)
                summary += "\n".join(
                    [
                        flow.section("info").string,
                        ",".join(flow.match.keys()),
                        "actions: " + ",".join(list(a.keys())[0] for a in flow.actions),
                    ]
                )
                attr = (
                    node_styles.get(
                        next(filter(lambda f: f.evaluate(flow), node_styles), None)
                    )
                    or {}
                )

                sg.node(
                    name=name,
                    label=summary,
                    _attributes=attr,
                    fontsize="8",
                    nojustify="true",
                    URL="#flow_{}".format(flow.id),
                )

                if previous:
                    sg.edge(previous, name, color="white")
                else:
                    sg.edge(invis, name, color="white", length="0")
                previous = name

                next_recirc = next(
                    (kv.value for kv in flow.actions_kv if kv.key == "recirc"), None
                )
                if next_recirc:
                    cname = "cluster_{}".format(hex(next_recirc))
                    g.edge(name, "f{}".format(hex(next_recirc)), lhead=cname)
                else:
                    g.edge(name, "end")

    g.edge("start", "f0x0", lhead="cluster_0x0")
    g.node("start", shape="Mdiamond")
    g.node("end", shape="Msquare")

    if not html:
        print(g.source)
        return

    html_obj = ""
    html_obj += "<h1> Flow Graph </h1>"
    html_obj += "<div width=400px height=300px>"
    svg = g.pipe(format="svg")
    html_obj += svg.decode("utf-8")
    html_obj += "</div>"

    html_obj += get_html_obj(list(itertools.chain(*recirc_flows.values())), opts)
    print(html_obj)


class HTMLFlowTree:
    def __init__(self, flow=None, opts=None):
        self._flow = flow
        self._formatter = HTMLFormatter(opts)
        self._subflows = list()
        self._opts = opts

    def append(self, flow):
        self._subflows.append(flow)

    def render(self, item=0):
        html_obj = "<div>"
        if self._flow:
            html_obj += """
        <input id="collapsible_{item}" class="toggle" type="checkbox" onclick="toggle_checkbox(this)" checked>
        <label for="collapsible_{item}" class="lbl-toggle lbl-toggle-flow">Flow {id}</label>
        """.format(
                item=item, id=self._flow.id
            )
            html_obj += '<div class="flow collapsible-content" id="flow_{id}" onfocus="onFlowClick(this)" onclick="onFlowClick(this)" >'.format(
                id=self._flow.id
            )
            buf = HTMLBuffer()
            highlighted = None
            if self._opts.get("highlight"):
                result = self._opts.get("highlight").evaluate(self._flow)
                if result:
                    highlighted = result.kv
            self._formatter.format_flow(buf, self._flow, highlighted)
            html_obj += buf.text
            html_obj += "</div>"
        if self._subflows:
            html_obj += "<div>"
            html_obj += "<ul  style='list-style-type:none;'>"
            for sf in self._subflows:
                item += 1
                html_obj += "<li>"
                (html_elem, items) = sf.render(item)
                html_obj += html_elem
                item += items
                html_obj += "</li>"
            html_obj += "</ul>"
            html_obj += "</div>"
        html_obj += "</div>"
        return html_obj, item


def get_html_obj(flow_list, opts=None):
    def append_to_html(parent, flow):
        html_flow = HTMLFlowTree(flow, opts)
        parent.append(html_flow)
        return html_flow

    root = HTMLFlowTree(flow=None, opts=opts)
    process_flow_tree(flow_list, root, 0, append_to_html)

    html_obj = """
    <style>
    .flow{
        background-color:white;
        display: inline-block;
        text-align: left;
        font-family: monospace;
    }
    .active{
        border: 2px solid #0008ff;
    }
    input[type='checkbox'] { display: none; }
    .wrap-collabsible {
        margin: 1.2rem 0;
    }
    .lbl-toggle-main {
        font-weight: bold;
        font-family: monospace;
        font-size: 1.5rem;
        text-transform: uppercase;
        text-align: center;
        padding: 1rem;
        #cursor: pointer;
        border-radius: 7px;
        transition: all 0.25s ease-out;
    }
    .lbl-toggle-flow {
        font-family: monospace;
        font-size: 1.0rem;
        text-transform: uppercase;
        text-align: center;
        padding: 1rem;
        #cursor: pointer;
        border-radius: 7px;
        transition: all 0.25s ease-out;
    }
    .lbl-toggle:hover {
        color: #0008ff;
    }
    .lbl-toggle::before {
        content: ' ';
        display: inline-block;
        border-top: 5px solid transparent;
        border-bottom: 5px solid transparent;
        border-left: 5px solid currentColor;
        vertical-align: middle;
        margin-right: .7rem;
        transform: translateY(-2px);
        transition: transform .2s ease-out;
    }
    .toggle:checked+.lbl-toggle::before {
        transform: rotate(90deg) translateX(-3px);
    }
    .collapsible-content {
        max-height: 0px;
        overflow: hidden;
        transition: max-height .25s ease-in-out;
    }
    .toggle:checked + .lbl-toggle + .collapsible-content {
        max-height: 350px;
    }
    .toggle:checked+.lbl-toggle {
        border-bottom-right-radius: 0;
        border-bottom-left-radius: 0;
    }
    .collapsible-content .content-inner {
        background: rgba(0, 105, 255, .2);
        border-bottom: 1px solid rgba(0, 105, 255, .45);
        border-bottom-left-radius: 7px;
        border-bottom-right-radius: 7px;
        padding: .5rem 1rem;
    }
    .collapsible-content p {
        margin-bottom: 0;
    }
    </style>

    <script>
      function onFlowClick(elem) {
          var flows = document.getElementsByClassName("flow");
          for (i = 0; i < flows.length; i++) {
              flows[i].classList.remove('active')
          }
          elem.classList.add("active");
          var my_toggle = document.getElementsByClassName("flow");
          toggleAll(elem, true);
      }
      function locationHashChanged() {
          var elem = document.getElementById(location.hash.substring(1));
          console.log(elem)
          if (elem) {
            if (elem.classList.contains("flow")) {
                onFlowClick(elem);
            }
          }
      }
      function toggle_checkbox(elem) {
         if (elem.checked == true) {
            toggleAll(elem, true)
         } else {
            toggleAll(elem, false)
         }
      }
      function toggleAll(elem, value) {
          var subs = elem.parentElement.querySelectorAll(".toggle:not([id=" + CSS.escape(elem.id) + "])");
          console.log(subs);
          console.log(value);
          for (i = 0; i < subs.length; ++i) {
              subs[i].checked = value;
          }
      }
      window.onhashchange = locationHashChanged;
    </script>
    """
    html_obj += """
        <input id="collapsible_main" class="toggle" type="checkbox" onclick="toggle_checkbox(this)" checked>
        <label for="collapsible_main" class="lbl-toggle lbl-toggle-main">Flow Table</label>
        """
    html_obj += "<div id=flow_list>"
    (html_elem, items) = root.render()
    html_obj += html_elem
    html_obj += "</div>"
    return html_obj


def process_flow_tree(flow_list, parent, recirc_id, callback):
    """Process the datapath flows into a tree by "recirc_id" and sorted by "packets"
    Args:
        flow_list (list[odp.ODPFlow]): original list of flows
        parent (Any): current tree node that serves as parent
        recirc_id (int): recirc_id to traverse
        callback(callable): a callback that must accept the current parent and
            a flow and return an object that can potentially serve as parent for
            a nested call to callback

    This function is recursive
    """
    sorted_flows = sorted(
        filter(lambda f: f.match.get("recirc_id") == recirc_id, flow_list),
        key=lambda x: x.info.get("packets") or 0,
        reverse=True,
    )

    for flow in sorted_flows:
        next_recirc = next(
            (kv.value for kv in flow.actions_kv if kv.key == "recirc"), None
        )

        next_parent = callback(parent, flow)

        if next_recirc:
            process_flow_tree(flow_list, next_parent, next_recirc, callback)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md") as readme_file:
    readme = readme_file.read()

requirements = ["click", "rich", "pyparsing", "netaddr", "graphviz"]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
]

setup(
    author="Adrián Moreno",
    author_email="amorenoz@redhat.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    description="OVS Debug contains scripts and libraries that help debug OVS"
    " and OVN",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    include_package_data=True,
    keywords="ovs_dbg",
    name="ovs_dbg",
    packages=find_packages(include=["ovs_dbg", "ovs_dbg.ofparse"]),
    setup_requires=setup_requirements,
    scripts=["bin/ofparse", "bin/ovs-lgrep"],
    data_files=["ovs_dbg/ofparse/ofparse.conf"],
    test_suite="tests",
    tests_require=test_requirements,
    url="https://ovs-dbg.readthedocs.io/en/latest/",
    project_urls={
        "Source": "https://github.com/amorenoz/ovs-dbg",
    },
    version="0.0.9",
    zip_safe=False,
)



""" OVS Log parser
"""

import datetime


class OVSLog:
    """Parses and stores information of a OVS FLow.

    The logs have the following standard format:

    2016-03-08T02:10:01.155Z|01417|vlog|INFO|opened log file /var/log/openvswitch/ovs-vswitchd.log
    2016-03-08T02:20:05.425Z|01418|connmgr|INFO|br0<->unix: 1 flow_mods in the last 0 s (1 adds)
    2016-03-08T02:20:10.160Z|01419|connmgr|INFO|br0<->unix: 1 flow_mods in the last 0 s (1 deletes)
    2016-03-08T11:30:52.206Z|00013|fatal_signal|WARN|terminating with signal 15 (Terminated)

    Attributes:
        timestamp: the UTC time stamp
        sequence: the sequence number of this message
        module: the module in vlog that emitted this error
        level: the level of error
        message: the rest of the message
        raw_message: the raw string
    """

    time_format = "%Y-%m-%dT%H:%M:%S.%fZ"
    _fields = ["timestamp", "sequence", "module", "level", "message"]

    def __init__(self, string):
        """Constructor"""
        self.raw_string = string
        fields = string.split("|", 4)
        if len(fields) != len(self._fields):
            raise Exception("Not a valid OVS Flow: %s" % string)

        self.timestamp = datetime.datetime.strptime(fields[0], self.time_format)
        self.sequence = int(fields[1])
        self.module = fields[2]
        self.level = fields[3]
        self.message = fields[4].strip()



""" Defines helpful decoders that can be used to decode information from the
flows

A decoder is generally a callable that accepts a string and returns the value
object
"""

import functools
import json
import netaddr
import re


class Decoder:
    """Base class for all decoder classes"""

    def to_json(self):
        assert "function must be implemented by derived class"


def decode_default(value):
    """Default decoder.

    It tries to convert into an integer value and, if it fails, just
    returns the string.
    """
    try:
        ival = int(value, 0)
        return ival
    except ValueError:
        return value


def decode_flag(value):
    """Default a flag. It's exising is just flagged by returning True"""
    return True


def decode_int(value):
    """integer decoder

    Both base10 and base16 integers are supported

    Used for fields such as:
        n_bytes=34
        metadata=0x4
    """
    return int(value, 0)


def decode_time(value):
    """time decoder

    Used for fields such as:
        duration=1234.123s
    """
    if value == "never":
        return value

    time_str = value.rstrip("s")
    return float(time_str)


class IntMask(Decoder):
    """Base class for Integer Mask decoder classes

    It supports decoding a value/mask pair. It has to be derived and size
    has to be set to a specific value
    """

    size = None  # size in bits

    def __init__(self, string):
        if not self.size:
            assert "IntMask should be derived and size should be fixed"

        parts = string.split("/")
        if len(parts) > 1:
            self._value = int(parts[0], 0)
            self._mask = int(parts[1], 0)
            if self._mask.bit_length() > self.size:
                raise ValueError(
                    "Integer mask {} is bigger than size {}".format(
                        self._mask, self.size
                    )
                )
        else:
            self._value = int(parts[0], 0)
            self._mask = 2 ** self.size - 1

        if self._value.bit_length() > self.size:
            raise ValueError(
                "Integer value {} is bigger than size {}".format(self._value, self.size)
            )

    @property
    def value(self):
        return self._value

    @property
    def mask(self):
        return self._mask

    def max_mask(self):
        return 2 ** self.size - 1

    def fully(self):
        """Returns if it's fully masked"""
        return self._mask == self.max_mask()

    def min(self):
        return self._value & self._mask

    def max(self):
        return (self.max_mask() & ~self._mask) | (self._value & self._mask)

    def __str__(self):
        if self.fully():
            return str(self._value)
        else:
            return "{}/{}".format(hex(self._value), hex(self._mask))

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self)

    def __eq__(self, other):
        return self.value == other.value and self.mask == other.mask

    def __contains__(self, other):
        if isinstance(other, IntMask):
            if other.fully():
                return other.value in self
            return other.min() in self and other.max() in self
        else:
            return other & self._mask == self._value & self._mask

    def dict(self):
        return {"value": self._value, "mask": self._mask}

    def to_json(self):
        return self.dict()


class Mask8(IntMask):
    size = 8


class Mask16(IntMask):
    size = 16


class Mask32(IntMask):
    size = 32


class Mask64(IntMask):
    size = 64


class Mask128(IntMask):
    size = 128


class Mask992(IntMask):
    size = 992


def decode_mask(mask_size):
    """value/mask decoder for values of specific size (bits)

    Used for fields such as:
        reg0=0x248/0xff
    """

    class Mask(IntMask):
        size = mask_size
        __name__ = "Mask{}".format(size)

    return Mask



class EthMask(Decoder):
    """EthMask represents an Ethernet address with optional mask

    It uses netaddr.EUI

    Attributes:
        eth (netaddr.EUI): the ethernet address
        mask (netaddr.EUI): Optional, the ethernet address mask

    Args:
        string (str): A string representing the masked ethernet address
            e.g: 00.11:22:33:44:55 or 01:00:22:00:33:00/01:00:00:00:00:00
    """

    def __init__(self, string):
        mask_parts = string.split("/")
        self._eth = netaddr.EUI(mask_parts[0])
        if len(mask_parts) == 2:
            self._mask = netaddr.EUI(mask_parts[1])
        else:
            self._mask = None

    @property
    def eth(self):
        """The ethernet address"""
        return self._eth

    @property
    def mask(self):
        """The ethernet address mask"""
        return self._mask

    def __eq__(self, other):
        """Returns True if this EthMask is numerically the same as other"""
        return self._mask == other._mask and self._eth == other._eth

    def __contains__(self, other):
        """
        Args:
            other (netaddr.EUI): another Ethernet address

        Returns:
            True if other falls into the masked address range
        """
        if isinstance(other, EthMask):
            if other._mask:
                raise ValueError("EthMask mask comparison not supported")
            return other._eth in self

        if self._mask:
            return (other.value & self._mask.value) == (
                self._eth.value & self._mask.value
            )
        else:
            return other == self._eth

    def __str__(self):
        if self._mask:
            return "/".join(
                [
                    self._eth.format(netaddr.mac_unix),
                    self._mask.format(netaddr.mac_unix),
                ]
            )
        else:
            return self._eth.format(netaddr.mac_unix)

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self)

    def to_json(self):
        return str(self)


class IPMask(Decoder):
    """IPMask stores an IPv6 or IPv4 and a mask

    It uses netaddr.IPAddress. The IPMask can be represented by a
    netaddr.IPNetwork (if it's a valid cidr) or by an ip and a mask

    Args:
        string (str): A string representing the ip/mask
    """

    def __init__(self, string):
        """Constructor"""
        self._ipnet = None
        self._ip = None
        self._mask = None
        try:
            self._ipnet = netaddr.IPNetwork(string)
        except netaddr.AddrFormatError:
            pass

        if not self._ipnet:
            # It's not a valid CIDR. Store ip and mask indendently
            parts = string.split("/")
            if len(parts) != 2:
                raise ValueError(
                    "value {}: is not an ipv4 or ipv6 address".format(string)
                )
            try:
                self._ip = netaddr.IPAddress(parts[0])
                self._mask = netaddr.IPAddress(parts[1])
            except netaddr.AddrFormatError as exc:
                raise ValueError(
                    "value {}: is not an ipv4 or ipv6 address".format(string)
                ) from exc

    def __eq__(self, other):
        """Returns True if this IPMask is numerically the same as other"""
        if isinstance(other, netaddr.IPNetwork):
            return self._ipnet and self._ipnet == other
        if isinstance(other, netaddr.IPAddress):
            return self._ipnet and self._ipnet.ip == other
        elif isinstance(other, IPMask):
            if self._ipnet:
                return self._ipnet == other._ipnet

            return self._ip == other._ip and self._mask == other._mask
        else:
            return False

    def __contains__(self, other):
        """
        Args:
            other (netaddr.IPAddres): another IP address

        Returns:
            True if other falls into the masked ip range
        """
        if isinstance(other, IPMask):
            if not other._ipnet:
                raise ValueError("ip/mask comparisons not supported")

            return (
                netaddr.IPAddress(other._ipnet.first) in self
                and netaddr.IPAddress(other._ipnet.last) in self
            )

        elif isinstance(other, netaddr.IPAddress):
            if self._ipnet:
                return other in self._ipnet
            return (other & self._mask) == (self._ip & self._mask)

    def cidr(self):
        """
        Returns True if the IPMask is a valid CIDR
        """
        return self._ipnet is not None

    @property
    def ip(self):
        """The IP address"""
        if self._ipnet:
            return self._ipnet.ip
        return self._ip

    @property
    def mask(self):
        """The IP mask"""
        if self._ipnet:
            return self._ipnet.netmask
        return self._mask

    def __str__(self):
        if self._ipnet:
            return str(self._ipnet)
        return "/".join([str(self._ip), str(self._mask)])

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self)

    def to_json(self):
        return str(self)


def decode_free_output(value):
    """Decodes the output value when found free (without the 'output' keyword)"""
    try:
        return "output", {"port": int(value)}
    except ValueError:
        return "output", {"port": value.strip('"')}


ipv4 = r"[\d\.]+"
ipv4_capture = r"({ipv4})".format(ipv4=ipv4)
ipv6 = r"[\w:]+"
ipv6_capture = r"(?:\[*)?({ipv6})(?:\]*)?".format(ipv6=ipv6)
port_range = r":(\d+)(?:-(\d+))?"
ip_range_regexp = r"{ip_cap}(?:-{ip_cap})?(?:{port_range})?"
ipv4_port_regex = re.compile(
    ip_range_regexp.format(ip_cap=ipv4_capture, port_range=port_range)
)
ipv6_port_regex = re.compile(
    ip_range_regexp.format(ip_cap=ipv6_capture, port_range=port_range)
)


def decode_ip_port_range(value):
    """
    Decodes an IP and port range:
        {ip_start}-{ip-end}:{port_start}-{port_end}

    IPv6 addresses are surrounded by "[" and "]" if port ranges are also
    present

    Returns the following dictionary:
        {
            "addrs": {
                "start": {ip_start}
                "end": {ip_end}
            }
            "ports": {
                "start": {port_start},
                "end": {port_end}
        }
        (the "ports" key might be omitted)
    """
    if value.count(":") > 1:
        match = ipv6_port_regex.match(value)
    else:
        match = ipv4_port_regex.match(value)

    ip_start = match.group(1)
    ip_end = match.group(2)
    port_start = match.group(3)
    port_end = match.group(4)

    result = {
        "addrs": {
            "start": netaddr.IPAddress(ip_start),
            "end": netaddr.IPAddress(ip_end or ip_start),
        }
    }
    if port_start:
        result["ports"] = {"start": int(port_start), "end": int(port_end or port_start)}

    return result


def decode_nat(value):
    """Decodes the 'nat' keyword of the ct action"""
    if not value:
        return True

    result = dict()
    type_parts = value.split("=")
    result["type"] = type_parts[0]

    if len(type_parts) > 1:
        value_parts = type_parts[1].split(",")
        if len(type_parts) != 2:
            raise ValueError("Malformed nat action: %s" % value)

        ip_port_range = decode_ip_port_range(value_parts[0])

        result = {"type": type_parts[0], **ip_port_range}

        for flag in value_parts[1:]:
            result[flag] = True

    return result


class FlowEncoder(json.JSONEncoder):
    """FlowEncoder is a json.JSONEncoder instance that can be used to
    serialize flow fields
    """

    def default(self, obj):
        if isinstance(obj, Decoder):
            return obj.to_json()
        elif isinstance(obj, netaddr.IPAddress):
            return str(obj)

        return json.JSONEncoder.default(self, obj)



""" Common helper classes for flow (ofproto/dpif) parsing
"""

import re
import functools

from ovs_dbg.decoders import decode_default


class ParseError(RuntimeError):
    """Exception raised when an error occurs during parsing."""

    pass


class KeyMetadata:
    """Class for keeping key metadata.

    Attributes:
        kpos (int): The position of the keyword in the parent string.
        vpos (int): The position of the value in the parent string.
        kstring (string): The keyword string as found in the flow string.
        vstring (string): The value as found in the flow string.
        end_del (bool): Whether the key has end delimiter.
    """

    def __init__(self, kpos, vpos, kstring, vstring, delim="", end_delim=""):
        """Constructor"""
        self.kpos = kpos
        self.vpos = vpos
        self.kstring = kstring
        self.vstring = vstring
        self.delim = delim
        self.end_delim = end_delim

    def __str__(self):
        return "key: [{},{}), val:[{}, {})".format(
            self.kpos,
            self.kpos + len(self.kstring),
            self.vpos,
            self.vpos + len(self.vstring),
        )

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self)


class KeyValue:
    """Class for keeping key-value data

    Attributes:
        key (str): The key string.
        value (any): The value data.
        meta (KeyMetadata): The key metadata.
    """

    def __init__(self, key, value, meta=None):
        """Constructor"""
        self.key = key
        self.value = value
        self.meta = meta

    def __str__(self):
        return "{}: {} ({})".format(self.key, str(self.value), str(self.meta))

    def __repr__(self):
        return "%s('%s')" % (self.__class__.__name__, self)


class KVDecoders:
    """KVDecoders class is used by KVParser to select how to decoode the value
    of a specific keyword.

    A decoder is simply a function that accepts a value string
    and returns the value objects to be stored.
    The returned value may be of any type.

    Decoders may return a KeyValue instance to indicate that the keyword should
    also be modified to match the one provided in the returned KeyValue

    The free_decoder, however, must return the key and value to be stored

    Args:
        decoders (dict): Optional; A dictionary of decoders indexed by keyword.
        default (callable): Optional; A decoder used if a match is not found in
            configured decoders. If not provided, the default behavior is to
            try to decode the value into an integer and, if that fails,
            just return the string as-is.
        default_free (callable): Optional; The decoder used if a match is not
            found in configured decoders and it's a free value (e.g:
            a value without a key) Defaults to returning the free value as
            keyword and "True" as value.
            The callable must accept a string and return a key, value pair
    """

    def __init__(self, decoders=None, default=None, default_free=None):
        self._decoders = decoders or dict()
        self._default = default or decode_default
        self._default_free = default_free or self._default_free_decoder

    def decode(self, keyword, value_str):
        """Decode a keyword and value.

        Args:
            keyword (str): The keyword whose value is to be decoded.
            value_str (str): The value string.

        Returns:
            The key (str) and value(any) to be stored.
        """

        decoder = self._decoders.get(keyword)
        if decoder:
            result = decoder(value_str)
            if isinstance(result, KeyValue):
                keyword = result.key
                value = result.value
            else:
                value = result

            return keyword, value
        else:
            if value_str:
                return keyword, self._default(value_str)
            else:
                return self._default_free(keyword)

    @staticmethod
    def _default_free_decoder(key):
        """Default decoder for free kewords."""
        return key, True


class KVParser:
    """KVParser parses a string looking for key-value pairs.

    Args:
        decoders (KVDecoders): Optional; the KVDecoders instance to use.
    """

    def __init__(self, decoders=None):
        """Constructor"""
        self._decoders = decoders or KVDecoders()
        self._keyval = list()

    def keys(self):
        return list(kv.key for kv in self._keyval)

    def kv(self):
        return self._keyval

    def __iter__(self):
        return iter(self._keyval)

    def parse(self, string):
        """Parse the key-value pairs in string.

        Args:
            string (str): the string to parse.

        Raises:
            ParseError if any parsing error occurs.
        """
        kpos = 0
        while kpos < len(string) and string[kpos] != "\n":
            # strip string
            if string[kpos] == "," or string[kpos] == " ":
                kpos += 1
                continue

            split_parts = re.split(r"(\(|=|:|,|\n|\r|\t|$)", string[kpos:], 1)
            # the delimiter should be included in the returned list
            if len(split_parts) < 3:
                break

            keyword = split_parts[0]
            delimiter = split_parts[1]
            rest = split_parts[2]

            value_str = ""
            vpos = kpos + len(keyword) + 1
            end_delimiter = ""

            # Figure out the end of the value
            # If the delimiter is ':' or '=', the end of the value is the end
            # of the string or a ', '
            if delimiter in ("=", ":"):
                value_parts = re.split(r"( |,|\n|\r|\t)", rest, 1)
                value_str = value_parts[0] if len(value_parts) == 3 else rest
                next_kpos = vpos + len(value_str)

            elif delimiter == "(":
                # Find the next ')'
                level = 1
                index = 0
                value_parts = re.split(r"(\(|\))", rest)
                for val in value_parts:
                    if val == "(":
                        level += 1
                    elif val == ")":
                        level -= 1
                    index += len(val)
                    if level == 0:
                        break

                if level != 0:
                    raise ParseError(
                        "Error parsing string {}: "
                        "Failed to find matching ')' in {}".format(string, rest)
                    )

                value_str = rest[: index - 1]
                next_kpos = vpos + len(value_str) + 1
                end_delimiter = ")"

                # Exceptionally, if after the () we find -> {}, do not treat
                # the content of the parenthesis as the value, consider
                # ({})->{} as the string value
                if index < len(rest) - 2 and rest[index : index + 2] == "->":
                    extra_val = rest[index + 2 :].split(",")[0]
                    value_str = "({})->{}".format(value_str, extra_val)
                    # remove the first "("
                    vpos -= 1
                    next_kpos = vpos + len(value_str)
                    end_delimiter = ""

            elif delimiter in (",", "\n", "\t", "\r", ""):
                # key with no value
                next_kpos = kpos + len(keyword)
                vpos = -1

            try:
                key, val = self._decoders.decode(keyword, value_str)
            except Exception as e:
                raise ParseError(
                    "Error parsing key-value ({}, {})".format(keyword, value_str)
                ) from e

            meta = KeyMetadata(
                kpos=kpos,
                vpos=vpos,
                kstring=keyword,
                vstring=value_str,
                delim=delimiter,
                end_delim=end_delimiter,
            )

            self._keyval.append(KeyValue(key, val, meta))

            kpos = next_kpos


def decode_nested_kv(decoders, value):
    """A key-value decoder that extracts nested key-value pairs and returns
    them in a dictionary

    Args:
        decoders (KVDecoders): the KVDecoders to use.
        value (str): the value string to decode.
    """
    if not value:
        # Mark as flag
        return True

    parser = KVParser(decoders)
    parser.parse(value)
    return {kv.key: kv.value for kv in parser.kv()}


def nested_kv_decoder(decoders=None):
    """Helper function that creates a nested kv decoder with given KVDecoders"""
    return functools.partial(decode_nested_kv, decoders)
